from cyclopts import App

from btpy.core.versions import list_bt_versions, get_bt_version

app = App()


@app.command(name="list")
async def list_versions_cmd():
    bt_versions = list_bt_versions()
    print([bt_version.version for bt_version in bt_versions])


@app.command(name="show")
async def show_version_cmd(version: str):
    bt_version = get_bt_version(version)
    print(bt_version)



from typing import Annotated

from cyclopts import App, Parameter

from btpy.core.ad import create_ad

app = App()


@app.command(name="test")
async def create_ad_cmd(
    id: Annotated[str, Parameter(name=["-i", "--id"], required=True)],
    name: Annotated[str, Parameter(name=["-n", "--name"], required=True)],
    resource_group: Annotated[
        str, Parameter(name=["-r", "--resource-group"], required=True)
    ],
    location: Annotated[str, Parameter(name=["-l", "--location"], required=True)],
    country_code: Annotated[
        str, Parameter(name=["-c", "--country-code"], required=True)
    ],
    subscription_id: Annotated[
        str, Parameter(name=["-s", "--subscription-id"], required=True)
    ],
):
    await create_ad(id, name, resource_group, location, country_code, subscription_id)



import json
import os
from dataclasses import dataclass, asdict
from datetime import datetime, timezone

import yaml

from btpy.configuration.config import StorageAccountConfig, get_config_value
from btpy.core.azure_utility import get_storage_table_entry, list_storage_table_entries

BT_VERSIONS_TABLE_NAME = "btversions"


@dataclass
class BtServiceVersion:
    version: str
    image: str | None


@dataclass
class BtVersion:
    version: str
    release_date: datetime | None
    services: list[BtServiceVersion]


@dataclass
class VersionFileData:
    version: str
    upgraded_from: str | None
    upgraded_at: str


def list_bt_versions() -> list[BtVersion]:
    storage_account_info = _get_storage_account_info()
    result = list_storage_table_entries(
        storage_account_info.name,
        BT_VERSIONS_TABLE_NAME,
        storage_account_info.subscription,
    )
    return [BtVersion(**json.loads(version["data"])) for version in result["items"]]


def get_bt_version(version) -> BtVersion | None:
    storage_account_info = _get_storage_account_info()
    result = get_storage_table_entry(
        storage_account_info.name,
        BT_VERSIONS_TABLE_NAME,
        storage_account_info.subscription,
        version,
        version,
    )
    if not result:
        return None
    return BtVersion(**json.loads(result["data"]))


def write_version_file(repo_path, new_version, old_version, overwrite=False):
    version_file_path = get_version_file_path(repo_path)
    if os.path.exists(version_file_path) and not overwrite:
        return
    version_data = VersionFileData(
        new_version, old_version, datetime.now(timezone.utc).isoformat()
    )
    with open(version_file_path, "w") as version_file:
        yaml.safe_dump(asdict(version_data), version_file)


def read_version_file(repo_path):
    version_file_path = get_version_file_path(repo_path)
    with open(version_file_path, "r") as version_file:
        return VersionFileData(**yaml.safe_load(version_file))


def get_version_file_path(repo_path):
    return os.path.join(repo_path, ".bt-version")


def _get_storage_account_info():
    storage_account_info = StorageAccountConfig(
        **get_config_value("version_storage_account")
    )

    if not storage_account_info:
        raise Exception("Version storage account config not found")

    return storage_account_info



import os
import sys


def check_if_script_dir_exists():
    """
    Check if the script is being run from the root of the terraform entity.
    All scripts are expected to have the terraform main.tf file in
    the current working directory.

    return the name of the current directory which is the name of the current feature
    """
    if not os.path.exists(os.path.join(os.getcwd(), "scripts")):
        print("Please run this script from the root of the repository.")
        sys.exit(1)
    else:
        return os.getcwd().split(os.sep)[-1].strip().replace(" ", "_")



import os

from cyclopts import App

from btpy.configuration.config import get_config_path, write_default_config

app = App()


@app.command(name="reset")
async def reset_config_cmd():
    config_path = get_config_path()
    write_default_config(os.path.join(config_path, "config.yaml"))



from rich import print
from btpy.core.azure_utility import (
    create_azure_ad,
    get_azure_ad,
    get_azure_ad_app_by_name,
    create_azure_ad_app,
    create_azure_ad_app_secret,
    get_azure_ad_app_secret_by_name,
)


async def create_ad(
    id,
    name,
    resource_group,
    location,
    country_code,
    subscription_id,
    redirect_uris=None,
):
    """Create Azure AD B2C tenant with app registration. User flows must be created manually."""
    if redirect_uris is None:
        redirect_uris = []

    tenant_domain = f"{id}.onmicrosoft.com"
    ad_info = get_azure_ad(tenant_domain, resource_group, subscription_id)
    if not ad_info:
        print(f"[blue]Creating Azure AD {id}...", end=" ")
        ad_info = create_azure_ad(
            id, name, resource_group, location, country_code, subscription_id
        )
        print("[green]Complete")
    else:
        print(f"[blue]Azure AD {id} already exists")

    app_name = "beacontower"
    app = await get_azure_ad_app_by_name(app_name, ad_info.tenant_id)
    if not app:
        print(f"[blue]Creating Azure AD app registration {app_name}...", end=" ")
        app = await create_azure_ad_app(app_name, redirect_uris, ad_info.tenant_id)
        print("[green]Complete")
    else:
        print(f"[blue]Azure AD app registration {app_name} already exists")

    secret_name = "btpy-secret"
    secret = await get_azure_ad_app_secret_by_name(
        app.id, secret_name, ad_info.tenant_id
    )
    if not secret:
        print(
            f"[blue]Creating Azure AD app registration secret {secret_name}...", end=" "
        )
        secret = await create_azure_ad_app_secret(
            app.id, secret_name, ad_info.tenant_id
        )
        print("[green]Complete")
        print(f"[yellow]Secret value (save this!):[/yellow] {secret.secret_text}")
    else:
        print(f"[blue]Azure AD app registration secret {secret_name} already exists")

    print()
    print("[yellow]Manual step required:[/yellow] Create user flow in Azure Portal:")
    print("  1. Go to: https://portal.azure.com")
    print(f"  2. Switch to directory: [cyan]{tenant_domain}[/cyan]")
    print("  3. Search for 'Azure AD B2C' -> User flows -> New user flow")
    print("  4. Select 'Sign up and sign in' -> Recommended -> Create")

    return ad_info



import os
import sys
from dataclasses import dataclass
from pathlib import Path

import yaml

ENV_DATA_STORAGE_NAME_ENV_VAR_KEY = "BTPY_ENV_DATA_STORAGE_NAME"
ENV_DATA_STORAGE_NAME_CONFIG_KEY = "EnvDataStorageName"


@dataclass
class KeyVaultInfo:
    name: str
    tenant: str
    subscription: str


@dataclass
class RancherMgmtConfig:
    url: str


@dataclass
class DomainConfig:
    default: str


@dataclass
class StorageAccountConfig:
    name: str
    tenant: str
    subscription: str


@dataclass
class BtpyConfig:
    domain: DomainConfig
    EnvDataStorageName: str
    env_list_storage_account: KeyVaultInfo
    global_keyvault: KeyVaultInfo
    rancher_mgmt: RancherMgmtConfig
    version_storage_account: StorageAccountConfig


def get_config_path() -> Path:
    if os.name == "nt":  # Windows
        return Path(os.getenv("APPDATA", Path.home() / "AppData" / "Roaming")) / "btpy"
    elif os.name == "posix":  # Unix-like systems (Linux, macOS, etc.)
        if sys.platform == "darwin":  # macOS
            return Path.home() / "Library" / "Application Support" / "btpy"
        else:  # Linux and other Unix-like systems
            return Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config")) / "btpy"
    else:
        raise NotImplementedError(f"Unsupported platform: {os.name}")


def get_config_value(key):
    config_path = get_config_path()
    config_file_path = os.path.expanduser(os.path.join(config_path, "config.yaml"))

    if not os.path.exists(config_file_path):
        write_default_config(config_file_path)

    with open(config_file_path, "r") as config_file:
        config_yaml = yaml.safe_load(config_file)
        return config_yaml.get(key)


def get_config() -> BtpyConfig:
    """Load the full typed config"""
    config_path = get_config_path()
    config_file_path = os.path.expanduser(os.path.join(config_path, "config.yaml"))

    if not os.path.exists(config_file_path):
        write_default_config(config_file_path)

    with open(config_file_path, "r") as config_file:
        config_yaml = yaml.safe_load(config_file)
        return BtpyConfig(
            EnvDataStorageName=config_yaml["EnvDataStorageName"],
            env_list_storage_account=KeyVaultInfo(
                **config_yaml["env_list_storage_account"]
            ),
            global_keyvault=KeyVaultInfo(**config_yaml["global_keyvault"]),
            rancher_mgmt=RancherMgmtConfig(**config_yaml["rancher_mgmt"]),
            domain=DomainConfig(**config_yaml["domain"]),
            version_storage_account=StorageAccountConfig(
                **config_yaml["version_storage_account"]
            ),
        )


def get_env_data_storage_name():
    # Try env var
    env_value = os.getenv(ENV_DATA_STORAGE_NAME_ENV_VAR_KEY)

    if env_value:
        return env_value

    # Try user config file
    conf = get_config_path()
    # conf = get_config_path().mkdir(parents=True, exist_ok=True)
    config_file_path = os.path.expanduser(os.path.join(conf, "config.yaml"))

    if not os.path.exists(config_file_path):
        raise FileNotFoundError(
            f"Failed to read env data storage name from file: {config_file_path} \
        (can also be set through env var: {ENV_DATA_STORAGE_NAME_ENV_VAR_KEY})"
        )

    with open(config_file_path, "r") as config_file:
        config_yaml = yaml.safe_load(config_file)
        config_value = config_yaml[ENV_DATA_STORAGE_NAME_CONFIG_KEY]

        if config_value:
            return config_value
        return None


def write_default_config(path):
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as config_file:
        config_file.write(
            """
EnvDataStorageName: stgbttfdata
env_list_storage_account:
  tenant: 953ee0f9-5c0c-4769-b099-eca3c4193550
  subscription: bf222aeb-b358-4dc5-b2a3-07503773a0e3
  name: stgbttfdata
global_keyvault:
  tenant: 953ee0f9-5c0c-4769-b099-eca3c4193550
  subscription: bf222aeb-b358-4dc5-b2a3-07503773a0e3
  name: btdeploymentdata
version_storage_account:
  tenant: 953ee0f9-5c0c-4769-b099-eca3c4193550
  subscription: bf222aeb-b358-4dc5-b2a3-07503773a0e3
  name: stgbttfdata
rancher_mgmt:
  url: https://bt-rancher.northeurope.cloudapp.azure.com/v3
domain:
  default: beacontower.app
"""
        )



import importlib.metadata

from cyclopts import App

from btpy.commands import (
    config_commands,
    env_commands,
    iac_commands,
    migration_commands,
    software_commands,
    tenant_commands,
    version_commands,
    ad_commands,
)


def get_version():
    return importlib.metadata.version("beacontower-btpy")


app = App(version=get_version)

app.command(env_commands.app, name="env")
app.command(software_commands.app, name="software")
app.command(iac_commands.app, name="iac")
app.command(migration_commands.app, name="migration")
app.command(tenant_commands.app, name="tenant")
app.command(config_commands.app, name="config")
app.command(version_commands.app, name="version")
app.command(ad_commands.app, name="ad")

if __name__ == "__main__":
    app()



import json
import platform
import shlex
import subprocess


def az(command: str, subscription_id: str = None):
    args = ["az", *shlex.split(command), "--output", "json"]

    if subscription_id:
        args.extend(["--subscription", subscription_id])

    program = subprocess.run(
        args, capture_output=True, text=True, shell=(platform.system() == "Windows")
    )

    stdout_data = None
    if program.stdout and program.stdout.strip():
        try:
            stdout_data = json.loads(program.stdout)
        except json.JSONDecodeError:
            stdout_data = program.stdout

    return program.returncode, stdout_data, program.stderr



from typing import Annotated

from cyclopts import App, Parameter
from rich import print
from rich.table import Table

from btpy.configuration.aliases import (
    load_aliases,
    remove_alias,
    resolve_env_alias,
    resolve_tenant_alias,
    set_alias,
)
from btpy.core.tenant.tenant import (
    TenantCreationOptions,
    create_tenant,
    list_tenants,
    register_tenant_in_env,
    unregister_tenant_from_env,
    get_tenant_debug_telemetry_config,
    set_tenant_debug_telemetry_loglevel,
    listen_to_tenant_eventhub,
    set_tenant_debug_telemetry_tracing,
    get_tenant_login_info,
)

app = App()


@app.command(name="create")
async def create_tenant_cmd(
    *,
    customer: Annotated[str, Parameter(name=["-c", "--customer"], required=True)],
    tenant_env: Annotated[str, Parameter(name=["-t", "--tenant-env"], required=True)],
    location: Annotated[str, Parameter(name=["-l", "--location"], required=True)],
    cluster_env: Annotated[
        str, Parameter(name=["-e", "--environment"], required=True)
    ] = None,
    version: Annotated[
        str, Parameter(name=["-v", "--version"], required=False)
    ] = "latest",
    subdomain: Annotated[
        str, Parameter(name=["-d", "--subdomain"], required=False)
    ] = None,
    suffix: Annotated[str, Parameter(name=["-s", "--suffix"])] = None,
    regen_ssh: Annotated[bool, Parameter(name=["--regen-ssh"])] = False,
):
    """Create a new tenant repo folder and necessary resources.

    Args:
        -c, --customer: Customer name or identifier.
        -t, --tenant-env: Tenant environment name (e.g., dev, test, prod).
        -l, --location: Cloud location/region (e.g., westeurope, eastus).
        -e, --environment: Cluster environment name or alias.
        -v, --version: Version to use for templates/resources. Defaults to "latest".
        -d, --subdomain: Subdomain used for the tenant ingress. Defaults to tenant name, e.g. "glaze-prod-neu1234".
        -s, --suffix: Optional suffix appended to generated resource names. Will be generated if unset.
        --regen-ssh: Regenerate SSH keys even if they already exist in Key Vault.
    """

    cluster_env = resolve_env_alias(cluster_env)

    options = TenantCreationOptions(
        customer,
        tenant_env,
        location,
        cluster_env,
        subdomain,
        version,
        suffix,
        regen_ssh,
    )
    create_tenant(options)


@app.command(name="list")
async def list_tenants_cmd(
    *,
    env: Annotated[str, Parameter(name=["-e", "--env"], required=True)],
    by_repo: Annotated[
        bool, Parameter(name=["-r", "--by-repo"], required=False)
    ] = False,
):
    """List registered tenants.

    Lists tenants for the specified environment. Use ``--by-repo`` to list
    tenants discovered from repository configuration instead of environment registration.

    Args:
        -e, --env: Environment name or alias to filter tenants by. Required.
        -r, --by-repo: If set, list tenants by repository instead of environment. Optional.
    """
    method = "env-list" if not by_repo else "repo"
    env_name = resolve_env_alias(env)
    tenants = list_tenants(env_name, method)
    table = Table()

    table.add_column("name")

    for tenant in tenants:
        table.add_row(tenant)

    print(table)


@app.command(name="register")
async def register_tenant_repo_cmd(tenant_name):
    """Registers tenant with environment and its flux system"""

    tenant_name = resolve_tenant_alias(tenant_name)
    register_tenant_in_env(tenant_name)


@app.command(name="unregister")
async def unregister_tenant_repo_cmd(tenant_name):
    """Unregisters tenant from environment and its flux system"""

    tenant_name = resolve_tenant_alias(tenant_name)
    unregister_tenant_from_env(tenant_name)


@app.command(name="log")
async def tenant_log_level_cmd(
    log_level: str = None,
    *,
    tenant: Annotated[str, Parameter(name=["-t", "--tenant"], required=True)],
    services: Annotated[list[str], Parameter(name=["-s", "--service"])] = None,
):
    tenant_name = resolve_tenant_alias(tenant)

    if not log_level:
        log_levels = get_tenant_debug_telemetry_config(tenant_name, services)
        table = Table()
        table.add_column("pod")
        table.add_column("log level")

        for pod in sorted(log_levels.items()):
            table.add_row(pod[0], pod[1]["logLevel"])

        print(table)
    else:
        log_levels = set_tenant_debug_telemetry_loglevel(
            tenant_name, log_level, services
        )
        table = Table()
        table.add_column("pod")
        table.add_column("log level")

        for pod in sorted(log_levels.items()):
            table.add_row(pod[0], pod[1]["logLevel"])

        print(table)


@app.command(name="trace")
async def tenant_trace_cmd(
    trace_enabled: bool = None,
    *,
    tenant: Annotated[str, Parameter(name=["-t", "--tenant"], required=True)],
    services: Annotated[list[str], Parameter(name=["-s", "--service"])] = None,
):
    tenant_name = resolve_tenant_alias(tenant)

    if trace_enabled is None:
        trace_settings = get_tenant_debug_telemetry_config(tenant_name, services)
        table = Table()
        table.add_column("pod")
        table.add_column("tracing")

        for pod in sorted(trace_settings.items()):
            table.add_row(pod[0], pod[1]["tracingEnabled"])

        print(table)
    else:
        trace_settings = set_tenant_debug_telemetry_tracing(
            tenant_name, trace_enabled, services
        )
        table = Table()
        table.add_column("pod")
        table.add_column("tracing")

        for pod in sorted(trace_settings.items()):
            table.add_row(pod[0], str(pod[1]["tracingEnabled"]))

        print(table)


@app.command(name="eh-listen")
async def tenant_listen_eh_cmd(
    eh_name: str,
    *,
    tenant: Annotated[str, Parameter(name=["-t", "--tenant"], required=True)],
):
    """Listen to event hub events

    Args:
        eh_name: Event hub name (without `eh-` prefix), e.g. device, ingress, orleans or timeseries.
    """
    tenant_name = resolve_tenant_alias(tenant)

    listen_to_tenant_eventhub(tenant_name, eh_name)


@app.command(name="login")
async def tenant_login_info_cmd(
    *, tenant: Annotated[str, Parameter(name=["-t", "--tenant"], required=True)]
):
    tenant_name = resolve_tenant_alias(tenant)
    login_info = get_tenant_login_info(tenant_name)
    print(login_info.url)
    print(login_info.email)
    print(login_info.password)


alias_app = App(name="alias", help="Manage tenant name aliases")


@alias_app.command(name="set")
def alias_set_cmd(alias: str, tenant_name: str):
    """Set an alias for a tenant name.

    Args:
        alias: Short name to use (e.g., 'prod-tenant')
        tenant_name: Full tenant name or keyvault name
    """
    set_alias(alias, tenant_name, alias_type="tenant")
    print(f"[green]Tenant alias '{alias}' -> '{tenant_name}' set successfully")


@alias_app.command(name="list")
def alias_list_cmd():
    """List all configured tenant aliases."""
    aliases = load_aliases(alias_type="tenant")

    if not aliases:
        print("[yellow]No tenant aliases configured")
        return

    table = Table(title="Tenant Aliases")
    table.add_column("Alias", style="cyan")
    table.add_column("Tenant Name", style="blue")

    for alias, tenant_name in sorted(aliases.items()):
        table.add_row(alias, tenant_name)

    print(table)


@alias_app.command(name="remove")
def alias_remove_cmd(alias: str):
    """Remove a tenant alias.

    Args:
        alias: The alias to remove
    """
    if remove_alias(alias, alias_type="tenant"):
        print(f"[green]Tenant alias '{alias}' removed successfully")
    else:
        print(f"[yellow]Tenant alias '{alias}' not found")


app.command(alias_app)



# Rancher Python Client
# Source: https://github.com/rancher/client-python
# License: MIT
# Internalized to avoid git dependency for PyPI publishing

import collections
import hashlib
import json
import os
import re
import time

import requests

PREFIX = "BTPY"
CACHE_DIR = "~/." + PREFIX.lower()
TIME = os.environ.get("TIME_API") is not None
DEFAULT_TIMEOUT = 45

LIST = "list-"
CREATE = "create-"
UPDATE = "update-"
DELETE = "delete-"
ACTION = "action-"
TRIM = True
JSON = False

GET_METHOD = "GET"
POST_METHOD = "POST"
PUT_METHOD = "PUT"
DELETE_METHOD = "DELETE"

HEADERS = {"Accept": "application/json"}

LIST_METHODS = {"__iter__": True, "__len__": True, "__getitem__": True}


def convert_type_name(type):
    if isinstance(type, str) is True:
        return type.replace(".", "_").replace("-", "")
    return type


def echo(fn):
    def wrapped(*args, **kw):
        ret = fn(*args, **kw)
        print(fn.__name__, repr(ret))
        return ret

    return wrapped


def timed_url(fn):
    def wrapped(*args, **kw):
        if TIME:
            start = time.time()
            ret = fn(*args, **kw)
            delta = time.time() - start
            print(delta, args[1], fn.__name__)
            return ret
        else:
            return fn(*args, **kw)

    return wrapped


class RestObject:
    def __init__(self):
        pass

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        data = {}
        for k, v in self.__dict__.items():
            if self._is_public(k, v):
                data[k] = v
        return repr(data)

    def __getattr__(self, k):
        if self._is_list() and k in LIST_METHODS:
            return getattr(self.data, k)
        return getattr(self.__dict__, k)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        if self._is_list():
            return iter(self.data)
        else:
            data = {}
            for k, v in self.__dict__.items():
                if self._is_public(k, v):
                    data[k] = v
            return iter(data.keys())

    def __len__(self):
        if self._is_list():
            return len(self.data)
        else:
            data = {}
            for k, v in self.__dict__.items():
                if self._is_public(k, v):
                    data[k] = v
            return len(data)

    @staticmethod
    def _is_public(k, v):
        return not callable(v)

    def _is_list(self):
        return "data" in self.__dict__ and isinstance(self.data, list)

    def data_dict(self):
        data = {}
        for k, v in self.__dict__.items():
            if self._is_public(k, v):
                data[k] = v
        return data


class Schema:
    def __init__(self, text, obj):
        self.text = text
        self.types = {}
        for t in obj:
            if t.type != "schema":
                continue

            self.types[convert_type_name(t.id)] = t
            t.creatable = False
            try:
                if POST_METHOD in t.collectionMethods:
                    t.creatable = True
            except AttributeError:
                pass

            t.updatable = False
            try:
                if PUT_METHOD in t.resourceMethods:
                    t.updatable = True
            except AttributeError:
                pass

            t.deletable = False
            try:
                if DELETE_METHOD in t.resourceMethods:
                    t.deletable = True
            except AttributeError:
                pass

            t.listable = False
            try:
                if GET_METHOD in t.collectionMethods:
                    t.listable = True
            except AttributeError:
                pass

            if not hasattr(t, "collectionFilters"):
                t.collectionFilters = {}

    def __str__(self):
        return str(self.text)

    def __repr(self):
        return repr(self.text)


class ApiError(Exception):
    def __init__(self, obj):
        self.error = obj
        try:
            msg = "{} : {}\n\t{}".format(obj.code, obj.message, obj)
            super(ApiError, self).__init__(self, msg)
        except Exception:
            super(ApiError, self).__init__(self, "API Error")


class ClientApiError(Exception):
    pass


class Client:
    def __init__(
        self,
        access_key=None,
        secret_key=None,
        url=None,
        cache=False,
        cache_time=86400,
        strict=False,
        headers=None,
        token=None,
        verify=True,
        **kw,
    ):
        if verify == "False":
            verify = False
        self._headers = HEADERS.copy()
        if headers is not None:
            for k, v in headers.items():
                self._headers[k] = v
        if token is not None:
            self.token = token
            self._headers["Authorization"] = "Bearer " + token
        self._access_key = access_key
        self._secret_key = secret_key
        if self._access_key is None:
            self._auth = None
        else:
            self._auth = (self._access_key, self._secret_key)
        self._url = url
        self._cache = cache
        self._cache_time = cache_time
        self._strict = strict
        self.schema = None
        self._session = requests.Session()
        self._session.verify = verify

        if not self._cache_time:
            self._cache_time = 60 * 60 * 24  # 24 Hours

        self._load_schemas()

    def valid(self):
        return self._url is not None and self.schema is not None

    def object_hook(self, obj):
        if isinstance(obj, list):
            return [self.object_hook(x) for x in obj]

        if isinstance(obj, dict):
            result = RestObject()

            for k, v in obj.items():
                setattr(result, k, self.object_hook(v))

            for link in ["next", "prev"]:
                try:
                    url = getattr(result.pagination, link)
                    if url is not None:
                        setattr(result, link, lambda url=url: self._get(url))
                except AttributeError:
                    pass

            if hasattr(result, "type") and isinstance(getattr(result, "type"), str):
                if hasattr(result, "links"):
                    for link_name, link in result.links.items():

                        def cb_link(_link=link, **kw):
                            return self._get(_link, data=kw)

                        if hasattr(result, link_name):
                            setattr(result, link_name + "_link", cb_link)
                        else:
                            setattr(result, link_name, cb_link)

                if hasattr(result, "actions"):
                    for link_name, link in result.actions.items():

                        def cb_action(_link_name=link_name, _result=result, *args, **kw):
                            return self.action(_result, _link_name, *args, **kw)

                        if hasattr(result, link_name):
                            setattr(result, link_name + "_action", cb_action)
                        else:
                            setattr(result, link_name, cb_action)

            return result

        return obj

    def object_pairs_hook(self, pairs):
        ret = collections.OrderedDict()
        for k, v in pairs:
            ret[k] = v
        return self.object_hook(ret)

    def _get(self, url, data=None):
        return self._unmarshall(self._get_raw(url, data=data))

    def _error(self, text):
        raise ApiError(self._unmarshall(text))

    @timed_url
    def _get_raw(self, url, data=None):
        r = self._get_response(url, data)
        return r.text

    def _get_response(self, url, data=None):
        r = self._session.get(url, auth=self._auth, params=data, headers=self._headers)
        if r.status_code < 200 or r.status_code >= 300:
            self._error(r.text)

        return r

    @timed_url
    def _post(self, url, data=None):
        r = self._session.post(
            url, auth=self._auth, data=self._marshall(data), headers=self._headers
        )
        if r.status_code < 200 or r.status_code >= 300:
            self._error(r.text)

        return self._unmarshall(r.text)

    @timed_url
    def _put(self, url, data=None):
        r = self._session.put(
            url, auth=self._auth, data=self._marshall(data), headers=self._headers
        )
        if r.status_code < 200 or r.status_code >= 300:
            self._error(r.text)

        return self._unmarshall(r.text)

    @timed_url
    def _delete(self, url):
        r = self._session.delete(url, auth=self._auth, headers=self._headers)
        if r.status_code < 200 or r.status_code >= 300:
            self._error(r.text)

        return self._unmarshall(r.text)

    def _unmarshall(self, text):
        if text is None or text == "":
            return text
        obj = json.loads(
            text, object_hook=self.object_hook, object_pairs_hook=self.object_pairs_hook
        )
        return obj

    def _marshall(self, obj, indent=None, sort_keys=False):
        if obj is None:
            return None
        return json.dumps(self._to_dict(obj), indent=indent, sort_keys=True)

    def _load_schemas(self, force=False):
        if self.schema and not force:
            return

        schema_text = self._get_cached_schema()

        if force or not schema_text:
            response = self._get_response(self._url)
            schema_url = response.headers.get("X-API-Schemas")
            if schema_url is not None and self._url != schema_url:
                schema_text = self._get_raw(schema_url)
            else:
                schema_text = response.text
            self._cache_schema(schema_text)

        obj = self._unmarshall(schema_text)

        schema = Schema(schema_text, obj)

        if len(schema.types) > 0:
            self._bind_methods(schema)
            self.schema = schema

    def reload_schema(self):
        self._load_schemas(force=True)

    def by_id(self, type, id, **kw):
        id = str(id)
        type_name = convert_type_name(type)
        url = self.schema.types[type_name].links.collection
        if url.endswith("/"):
            url += id
        else:
            url = "/".join([url, id])
        try:
            return self._get(url, self._to_dict(**kw))
        except ApiError as e:
            if e.error.status == 404:
                return None
            else:
                raise e

    def update_by_id(self, type, id, *args, **kw):
        type_name = convert_type_name(type)
        url = self.schema.types[type_name].links.collection
        if url.endswith("/"):
            url = url + id
        else:
            url = "/".join([url, id])

        return self._put_and_retry(url, *args, **kw)

    def update(self, obj, *args, **kw):
        url = obj.links.self
        return self._put_and_retry(url, *args, **kw)

    def _put_and_retry(self, url, *args, **kw):
        retries = kw.get("retries", 3)
        for i in range(retries):
            try:
                return self._put(url, data=self._to_dict(*args, **kw))
            except ApiError as e:
                if i == retries - 1:
                    raise e
                if e.error.status == 409:
                    time.sleep(0.1)
                else:
                    raise e

    def _post_and_retry(self, url, *args, **kw):
        retries = kw.get("retries", 3)
        for i in range(retries):
            try:
                return self._post(url, data=self._to_dict(*args, **kw))
            except ApiError as e:
                if i == retries - 1:
                    raise e
                if e.error.status == 409:
                    time.sleep(0.1)
                else:
                    raise e

    def _validate_list(self, type, **kw):
        if not self._strict:
            return

        type_name = convert_type_name(type)
        collection_filters = self.schema.types[type_name].collectionFilters

        for k in kw:
            if hasattr(collection_filters, k):
                return

            for filter_name, filter_value in collection_filters.items():
                for m in filter_value.modifiers:
                    if k == "_".join([filter_name, m]):
                        return

            raise ClientApiError(k + " is not searchable field")

    def list(self, type, **kw):
        type_name = convert_type_name(type)
        if type_name not in self.schema.types:
            raise ClientApiError(type_name + " is not a valid type")

        self._validate_list(type_name, **kw)
        collection_url = self.schema.types[type_name].links.collection
        return self._get(collection_url, data=self._to_dict(**kw))

    def reload(self, obj):
        return self.by_id(obj.type, obj.id)

    def create(self, type, *args, **kw):
        type_name = convert_type_name(type)
        collection_url = self.schema.types[type_name].links.collection
        return self._post(collection_url, data=self._to_dict(*args, **kw))

    def delete(self, *args):
        for i in args:
            if isinstance(i, RestObject):
                return self._delete(i.links.self)

    def action(self, obj, action_name, *args, **kw):
        url = getattr(obj.actions, action_name)
        return self._post_and_retry(url, *args, **kw)

    def _is_list(self, obj):
        if isinstance(obj, list):
            return True

        if (
            isinstance(obj, RestObject)
            and "type" in obj.__dict__
            and obj.type == "collection"
        ):
            return True

        return False

    def _to_value(self, value):
        if isinstance(value, dict):
            ret = {}
            for k, v in value.items():
                ret[k] = self._to_value(v)
            return ret

        if isinstance(value, list):
            ret = []
            for v in value:
                ret.append(self._to_value(v))
            return ret

        if isinstance(value, RestObject):
            ret = {}
            for k, v in vars(value).items():
                if not k.startswith("_") and not isinstance(v, RestObject) and not callable(v):
                    ret[k] = self._to_value(v)
                elif not k.startswith("_") and isinstance(v, RestObject):
                    ret[k] = self._to_dict(v)
            return ret

        return value

    def _to_dict(self, *args, **kw):
        if len(kw) == 0 and len(args) == 1 and self._is_list(args[0]):
            ret = []
            for i in args[0]:
                ret.append(self._to_dict(i))
            return ret

        ret = {}

        for i in args:
            value = self._to_value(i)
            if isinstance(value, dict):
                for k, v in value.items():
                    ret[k] = v

        for k, v in kw.items():
            ret[k] = self._to_value(v)

        return ret

    @staticmethod
    def _type_name_variants(name):
        ret = [name]
        python_name = re.sub(r"([a-z])([A-Z])", r"\1_\2", name)
        if python_name != name:
            ret.append(python_name.lower())

        return ret

    def _bind_methods(self, schema):
        bindings = [
            ("list", "collectionMethods", GET_METHOD, self.list),
            ("by_id", "collectionMethods", GET_METHOD, self.by_id),
            ("update_by_id", "resourceMethods", PUT_METHOD, self.update_by_id),
            ("create", "collectionMethods", POST_METHOD, self.create),
        ]

        for type_name, typ in schema.types.items():
            for name_variant in self._type_name_variants(type_name):
                for method_name, type_collection, test_method, m in bindings:

                    def cb_bind(type_name=type_name, method=m):
                        def _cb(*args, **kw):
                            return method(type_name, *args, **kw)

                        return _cb

                    if test_method in getattr(typ, type_collection, []):
                        setattr(self, "_".join([method_name, name_variant]), cb_bind())

    def _get_schema_hash(self):
        h = hashlib.new("sha1")
        h.update(self._url.encode())
        if self._access_key is not None:
            h.update(self._access_key.encode())
        return h.hexdigest()

    def _get_cached_schema_file_name(self):
        if not self._cache:
            return None

        h = self._get_schema_hash()

        cachedir = os.path.expanduser(CACHE_DIR)
        if not cachedir:
            return None

        if not os.path.exists(cachedir):
            os.mkdir(cachedir)

        return os.path.join(cachedir, "schema-" + h + ".json")

    def _cache_schema(self, text):
        cached_schema = self._get_cached_schema_file_name()

        if not cached_schema:
            return None

        with open(cached_schema, "w") as f:
            f.write(text)

    def _get_cached_schema(self):
        if not self._cache:
            return None

        cached_schema = self._get_cached_schema_file_name()

        if not cached_schema:
            return None

        if os.path.exists(cached_schema):
            mod_time = os.path.getmtime(cached_schema)
            if time.time() - mod_time < self._cache_time:
                with open(cached_schema) as f:
                    data = f.read()
                return data

        return None

    def wait_success(self, obj, timeout=-1):
        obj = self.wait_transitioning(obj, timeout)
        if obj.transitioning != "no":
            raise ClientApiError(obj.transitioningMessage)
        return obj

    def wait_transitioning(self, obj, timeout=-1, sleep=0.01):
        timeout = _get_timeout(timeout)
        start = time.time()
        obj = self.reload(obj)
        while obj.transitioning == "yes":
            time.sleep(sleep)
            sleep *= 2
            if sleep > 2:
                sleep = 2
            obj = self.reload(obj)
            delta = time.time() - start
            if delta > timeout:
                msg = "Timeout waiting for [{}:{}] to be done after {} seconds"
                msg = msg.format(obj.type, obj.id, delta)
                raise Exception(msg)

        return obj


def _get_timeout(timeout):
    if timeout == -1:
        return DEFAULT_TIMEOUT
    return timeout



import json
import os
import secrets
import shutil
import stat
import subprocess
import tempfile
import time
from dataclasses import asdict, dataclass

import yaml
from azure.eventhub import EventHubConsumerClient
from rich import print

from btpy.core.azure_utility import (
    assign_keyvault_reader_role,
    create_service_principal,
    delete_keyvault_secret,
    delete_service_principal,
    get_current_subscription,
    get_keyvault_secret,
    get_keyvault_secret_to_file,
    get_service_principals_by_name,
    location_tag,
    query_keyvault_secrets,
    upload_keyvault_secret,
    upload_keyvault_secret_file,
    upsert_table_storage_entry,
)
from btpy.core.envs.env2 import get_env_info, get_env_kubectl, _get_env_list_info
from btpy.core.file_utility import generate_file_from_template, to_absolute_path
from btpy.core.iac_utility import bootstrap_tf_repo, get_keyvault_name
from btpy.core.k8s_utility import parallel_pod_curl
from btpy.core.misc_utility import (
    create_domain,
    generate_password,
    get_gh_btenvs_app_info,
    init_git_repo,
    setup_ssh_keys,
)
from btpy.core.versions import get_bt_version, write_version_file
from btpy.iac.suffix import generate

TENANT_LIST_TABLE_NAME = "tenants"


@dataclass
class TenantTemplateData:
    resource_group: str
    env_name: str
    customer: str
    suffix: str
    location: str
    shloc: str
    keyvault_name: str
    acr_name: str
    acr_password: str
    acr_registry: str
    kubeconfig_path: str
    keyvault_sp_id: str
    keyvault_sp_secret: str
    tenant_namespace: str
    github_app_id: str
    github_app_private_key_path: str
    github_installation_id: str
    az_tenant_id: str
    tenant_domain: str
    subscription_id: str
    provider_iothubs: str


def list_tenants(env, method="env-list"):
    env_info = get_env_info(env)

    if method == "env-list":
        tenant_secret_prefix = "tenant-"
        tenant_secrets = query_keyvault_secrets(
            env_info.keyvault_name,
            f"[?contains(name, '{tenant_secret_prefix}')].name",
            env_info.subscription_id,
        )
        tenants = [name.removeprefix("tenant-repo-ssh-key-") for name in tenant_secrets]
        return tenants
    elif method == "repo":
        result = subprocess.run(
            [
                "gh",
                "repo",
                "list",
                "beacontower",
                "--json",
                "name",
                "-q",
                """.[]|select(.name|startswith("tenant-"))""",
            ],
            capture_output=True,
        )

        if result.returncode != 0:
            print(result.stderr)
            print(result.stdout)
            print("[red]Failed to query gh CLI for beacontower tenant repos")
            return

        tenants = json.loads(result.stdout)
        print(tenants)


@dataclass
class TenantCreationOptions:
    customer: str
    tenant_env: str
    location: str
    cluster_env: str
    subdomain: str | None = None
    version: str = "latest"
    suffix: str | None = None
    regen_ssh: bool = False


def create_tenant(options: TenantCreationOptions):
    env = get_env_info(options.cluster_env)

    if not env:
        print(f"[red]Environment cluster {options.cluster_env} does not exist")
        return

    env_kv = env.keyvault_name
    version = get_bt_version(options.version)

    if not version:
        print(f"[red]BT version {options.version} not found")
        return

    if options.suffix is None:
        options.suffix = generate()["suffix"]

    shloc = location_tag(options.location)
    tenant_name = f"{options.customer}-{options.tenant_env}-{shloc}{options.suffix}"
    folder = f"tenant-{tenant_name}"
    tf_folder = os.path.join(folder, "terraform")
    k8s_folder = os.path.join(folder, "kubernetes")

    print(f"[cyan]Creating tenant {tenant_name}")

    # Clone the tenant base repo and remove the .git folder
    if not os.path.isdir(folder):

        def __on_rm_error(func, path, exc):
            os.chmod(path, stat.S_IWRITE)
            func(path)

        # TODO: Should use version tags eventually
        repo_url = "https://github.com/beacontower/tenant-template.git"
        subprocess.run(
            ["git", "clone", "--branch", "master", repo_url, folder], check=True
        )
        shutil.rmtree(os.path.join(folder, ".git"), onerror=__on_rm_error)
    else:
        print(f"[blue]Repo folder {folder} already exists")

    # Gather some env values
    acr_name = get_keyvault_secret(env_kv, "acr-username")
    acr_password = get_keyvault_secret(env_kv, "acr-password")
    acr_registry = get_keyvault_secret(env_kv, "acr-registry")
    otel_url = get_keyvault_secret(env_kv, "otel-url")
    az_sub = get_current_subscription()
    subscription_id = az_sub["id"]

    # Init tfvars data
    tf_data = bootstrap_tf_repo(
        options.customer,
        options.tenant_env,
        options.location,
        options.suffix,
        "tenant",
        tf_folder,
    )

    if tf_data is None:
        return

    # TODO: Find better way to make sure rbac has been refreshed, if it fails when run, just run again
    time.sleep(6)

    keyvault_name = tf_data.keyvault_name
    sp = _setup_eso_reader_sp(
        f"sp-bt-tenant-{tenant_name}-kvr",
        keyvault_name,
        tf_data.keyvault_id,
        tf_data.subscription_id,
    )

    # Add tenant to tenant list
    _add_tenant_to_storage(tenant_name, env.name, keyvault_name)

    # Set some secrets in tenant key vault
    secrets_folder_name = ".secrets"
    secrets_path = os.path.join(tf_folder, secrets_folder_name)
    if not os.path.exists(secrets_path):
        os.mkdir(secrets_path)

    kubeconfig_path = os.path.join(secrets_path, "kubeconfig.yaml")
    relative_kubeconfig_path = os.path.join(secrets_folder_name, "kubeconfig.yaml")
    get_keyvault_secret_to_file(env_kv, "kubeconfig", kubeconfig_path)
    upload_keyvault_secret_file(keyvault_name, "kubeconfig", kubeconfig_path)
    upload_keyvault_secret(keyvault_name, "otel-url", otel_url)

    setup_ssh_keys(
        keyvault_name,
        secrets_path,
        "Auto-generated BT tenant repo SSH keys",
        options.regen_ssh,
    )

    _setup_charts_ssh_key(env_kv, keyvault_name, secrets_path)

    upload_keyvault_secret(keyvault_name, "env-keyvault-name", env_kv)
    upload_keyvault_secret(keyvault_name, "env-name", options.cluster_env)
    upload_keyvault_secret(keyvault_name, "tenant-name", tenant_name)

    # Get github auth info and store in secret file
    gh_app_info = get_gh_btenvs_app_info()
    gh_app_key_filename = "gh_app_key.pem"
    gh_app_private_key_path = os.path.join(secrets_path, gh_app_key_filename)
    relative_gh_app_private_key_path = os.path.join(
        secrets_folder_name, gh_app_key_filename
    )
    with open(gh_app_private_key_path, "w") as key_file:
        key_file.write(gh_app_info.private_key)

    # Generate tenant's domain
    tenant_subdomain = options.subdomain if options.subdomain else tenant_name
    tenant_domain = create_domain(tenant_subdomain)
    upload_keyvault_secret(keyvault_name, "tenant-domain", tenant_domain)

    iothub_name = f"iot-bt-device-{tenant_name}"
    provider_iothubs_string = json.dumps(
        [
            {
                "name": iothub_name,
                "resource_group": tf_data.resource_group,
                "dedicated": True,
                "dedicated_id": "defaultIotHubProvider",
            }
        ]
    )

    tenant_data = TenantTemplateData(
        resource_group=tf_data.resource_group,
        env_name=options.tenant_env,
        customer=options.customer,
        suffix=options.suffix,
        location=options.location,
        shloc=shloc,
        keyvault_name=keyvault_name,
        acr_name=acr_name,
        acr_password=acr_password,
        acr_registry=acr_registry,
        kubeconfig_path=relative_kubeconfig_path,
        keyvault_sp_id=sp.client_id,
        keyvault_sp_secret=sp.client_secret,
        tenant_namespace=f"tenant-{tenant_name}",
        github_app_id=gh_app_info.app_id,
        github_app_private_key_path=relative_gh_app_private_key_path,
        github_installation_id=gh_app_info.installation_id,
        az_tenant_id=tf_data.tenant_id,
        tenant_domain=tenant_domain,
        subscription_id=subscription_id,
        provider_iothubs=provider_iothubs_string,
    )
    generate_file_from_template(
        to_absolute_path(__file__, "templates/tenant-terraform.tfvars.json.mustache"),
        os.path.join(tf_folder, "terraform.tfvars.json"),
        asdict(tenant_data),
    )

    # Prepare git repo
    if not init_git_repo(folder, "tenant"):
        return

    write_version_file(folder, options.version, None, overwrite=False)

    # Generate app config values
    _create_app_config_values(keyvault_name, acr_registry, acr_name, acr_password)

    # TODO: Write next instructions as output


@dataclass
class ServicePrincipal:
    client_id: str
    client_secret: str


def _setup_charts_ssh_key(env_keyvault, tenant_keyvault, secret_folder_path):
    """Copy charts repo SSH key from env Key Vault to tenant Key Vault."""
    key_path = os.path.join(secret_folder_path, "charts_repo_ssh_key")
    get_keyvault_secret_to_file(env_keyvault, "charts-repo-ssh-key", key_path)
    upload_keyvault_secret_file(tenant_keyvault, "charts-repo-ssh-key", key_path)


def _setup_eso_reader_sp(name, keyvault_name, keyvault_id, subscription_id):
    existing_client_id = get_keyvault_secret(keyvault_name, "keyvault-sp-id")
    existing_client_secret = get_keyvault_secret(keyvault_name, "keyvault-sp-secret")

    if existing_client_id and existing_client_secret:
        # Make sure existing SP is key vault reader
        assign_keyvault_reader_role(keyvault_id, existing_client_id, subscription_id)
        return ServicePrincipal(existing_client_id, existing_client_secret)

    # Clean up potentially existing SPs
    existing_sps = get_service_principals_by_name(name)

    if existing_sps:
        for sp in existing_sps:
            delete_service_principal(sp["id"])

    sp = create_service_principal(name)
    assign_keyvault_reader_role(keyvault_id, sp["appId"], subscription_id)

    upload_keyvault_secret(keyvault_name, "keyvault-sp-id", sp["appId"])
    upload_keyvault_secret(keyvault_name, "keyvault-sp-secret", sp["password"])

    return ServicePrincipal(sp["appId"], sp["password"])


def _create_app_config_values(keyvault_name, acr_registry, acr_username, acr_password):
    # ACR credentials for pulling container images
    upload_keyvault_secret(keyvault_name, "acr-url", acr_registry)
    upload_keyvault_secret(keyvault_name, "acr-username", acr_username)
    upload_keyvault_secret(keyvault_name, "acr-password", acr_password)

    _create_user_service_config_values(keyvault_name)
    _create_arangodb_config_values(keyvault_name)


def _create_user_service_config_values(keyvault_name):
    # Default user
    upload_keyvault_secret(
        keyvault_name, "user-service-default-user-email", "beacontower@admin.io"
    )
    user_service_default_user_password = generate_password()
    upload_keyvault_secret(
        keyvault_name,
        "user-service-default-user-password",
        user_service_default_user_password,
    )

    # SpiceDB values
    upload_keyvault_secret(
        keyvault_name, "spicedb-postgres-username", "spicedb"
    )  # NOTE: Username must match the owner specified in the helmrelease in tenant-template
    spicedb_postgres_password = generate_password(uppercase=False, symbols=None)
    upload_keyvault_secret(
        keyvault_name, "spicedb-postgres-password", spicedb_postgres_password
    )

    spicedb_token = secrets.token_urlsafe(32)
    upload_keyvault_secret(keyvault_name, "spicedb-token", spicedb_token)

    # Timescale values
    upload_keyvault_secret(keyvault_name, "timescale-username", "timescale")
    timescale_password = generate_password(uppercase=False, symbols=None)
    upload_keyvault_secret(keyvault_name, "timescale-password", timescale_password)

    # TODO: Remove these temporary values when b2c/entry can be set up automatically
    upload_keyvault_secret(
        keyvault_name,
        "b2c-authority",
        "https://btglobalweufa2r.b2clogin.com/btglobalweufa2r.onmicrosoft.com/B2C_1_btsign_in/v2.0/",
    )
    upload_keyvault_secret(
        keyvault_name,
        "b2c-authority-web-client",
        "https://btglobalweufa2r.b2clogin.com/btglobalweufa2r.onmicrosoft.com/B2C_1_btsign_in",
    )
    upload_keyvault_secret(
        keyvault_name, "b2c-tenant-domain", "btglobalweufa2r.onmicrosoft.com"
    )
    upload_keyvault_secret(
        keyvault_name, "b2c-known-authority", "btglobalweufa2r.b2clogin.com"
    )
    upload_keyvault_secret(
        keyvault_name, "b2c-app-client-id", "7f535b5e-65e9-4d61-a6e5-e114687e909f"
    )
    upload_keyvault_secret(
        keyvault_name,
        "b2c-app-client-secret",
        "XrW8Q~AlsQBp5UxPG9_ncTxf.S-nQypP5~H4jaVX",
    )
    upload_keyvault_secret(
        keyvault_name,
        "b2c-metadata-url",
        "https://btglobalweufa2r.b2clogin.com/btglobalweufa2r.onmicrosoft.com/B2C_1_btsign_in/v2.0/.well-known/openid-configuration",
    )


def _create_arangodb_config_values(keyvault_name):
    arangodb_password = generate_password(uppercase=False, symbols=None)
    upload_keyvault_secret(keyvault_name, "arangodb-username", "beacontower")
    upload_keyvault_secret(keyvault_name, "arangodb-password", arangodb_password)


@dataclass
class TenantInfo:
    tenant_id: str
    subscription_id: str
    name: str
    env_name: str
    keyvault_name: str


def _add_tenant_to_storage(tenant_name, env_name, keyvault_name):
    print("[blue]Adding tenant to global env list...", end=" ")
    env_table_info = _get_env_list_info()
    sub_info = get_current_subscription()
    tenant_info = TenantInfo(
        tenant_id=sub_info["tenantId"],
        subscription_id=sub_info["id"],
        name=tenant_name,
        env_name=env_name,
        keyvault_name=keyvault_name,
    )
    upsert_table_storage_entry(
        env_table_info.name,
        TENANT_LIST_TABLE_NAME,
        tenant_name,
        tenant_name,
        tenant_info,
        env_table_info.subscription_id,
    )


def keyvault_name_to_tenant_name(keyvault_name):
    return keyvault_name[4:]


def register_tenant_in_env(tenant_name):
    print("[blue]Reading values...", end=" ")
    tenant_keyvault_name = get_keyvault_name(tenant_name)
    env_name = get_keyvault_secret(tenant_keyvault_name, "env-name")
    env_info = get_env_info(env_name)
    env_keyvault_name = env_info.keyvault_name
    tenant_name = get_keyvault_secret(tenant_keyvault_name, "tenant-name")
    env_name = get_keyvault_secret(env_keyvault_name, "env-name")
    az_tenant_id = env_info.tenant_id
    tenant_domain = get_keyvault_secret(tenant_keyvault_name, "tenant-domain")

    print("[green]Complete")

    # Register tenant repo SSH key in env key vault
    with tempfile.TemporaryDirectory() as temp_folder:
        tenant_ssh_key_path = os.path.join(temp_folder, "repo-ssh-key")
        get_keyvault_secret_to_file(
            tenant_keyvault_name, "repo-ssh-key", tenant_ssh_key_path
        )
        tenant_key_name = env_tenant_ssh_key_name(tenant_name)
        upload_keyvault_secret_file(
            env_keyvault_name, tenant_key_name, tenant_ssh_key_path
        )

    # Create tenant namespace and keyvault-sp-credentials secret
    print("[blue]Creating tenant namespace and credentials secret...", end=" ")
    tenant_namespace = f"tenant-{tenant_name}"

    kubeconfig_content = get_env_kubectl(env_name)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as kc_file:
        kc_file.write(kubeconfig_content)
        kubeconfig_path = kc_file.name

    try:
        # Create namespace (ignore if exists)
        result = subprocess.run(
            [
                "kubectl",
                "--kubeconfig",
                kubeconfig_path,
                "create",
                "namespace",
                tenant_namespace,
            ],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0 and "already exists" not in result.stderr:
            print(f"[red]Failed to create namespace ({result.stderr})")
            return

        # Get SP credentials from tenant KV
        sp_client_id = get_keyvault_secret(tenant_keyvault_name, "keyvault-sp-id")
        sp_client_secret = get_keyvault_secret(
            tenant_keyvault_name, "keyvault-sp-secret"
        )

        # Create secret (delete first if exists to allow update)
        subprocess.run(
            [
                "kubectl",
                "--kubeconfig",
                kubeconfig_path,
                "delete",
                "secret",
                "keyvault-sp-credentials",
                "-n",
                tenant_namespace,
                "--ignore-not-found",
            ],
            capture_output=True,
            text=True,
        )

        result = subprocess.run(
            [
                "kubectl",
                "--kubeconfig",
                kubeconfig_path,
                "create",
                "secret",
                "generic",
                "keyvault-sp-credentials",
                f"--from-literal=client_id={sp_client_id}",
                f"--from-literal=client_secret={sp_client_secret}",
                "-n",
                tenant_namespace,
            ],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"[red]Failed to create secret ({result.stderr})")
            return
        print("[green]Complete")
    finally:
        os.unlink(kubeconfig_path)

    # Create tenant flux files in env repo
    with tempfile.TemporaryDirectory() as temp_folder:
        # Generate tenant file
        tenant_file_name = f"tenant-{tenant_name}.yaml"
        tenant_path = os.path.join(temp_folder, tenant_file_name)
        tenant_namespace = f"tenant-{tenant_name}"
        tenant_file_data = {
            "tenant_name": tenant_name,
            "tenant_namespace": tenant_namespace,
            "tenant_domain": tenant_domain,
            "az_tenant_id": az_tenant_id,
            "keyvault_name": tenant_keyvault_name,
        }

        generate_file_from_template(
            to_absolute_path(__file__, "templates/flux-tenant.yaml.mustache"),
            tenant_path,
            tenant_file_data,
        )

        # Clone env repo
        print("[blue]Cloning env repo...", end=" ")
        env_repo_name = f"env-{env_name}"
        env_repo_url = f"https://github.com/btenvs/{env_repo_name}.git"
        env_repo_folder = os.path.join(temp_folder, "repo")

        result = subprocess.run(
            ["git", "clone", env_repo_url, env_repo_folder],
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            print("[green]Complete")
        else:
            print(f"[red]Failed ({result.stderr})")
            return

        # Update tenant kustomization.yaml to include new tenant file
        print(
            "[blue]Add tenant flux .yaml file into tenants kustomization list...",
            end=" ",
        )
        env_tenant_folder = os.path.join(env_repo_folder, "kubernetes", "tenants")
        tenant_kustomization_path = os.path.join(
            env_tenant_folder, "kustomization.yaml"
        )
        with open(tenant_kustomization_path, "r") as tenants_kustomization_file:
            tenants_file = yaml.safe_load(tenants_kustomization_file)

        if tenant_file_name not in tenants_file["resources"]:
            tenants_file["resources"].append(tenant_file_name)

            with open(tenant_kustomization_path, "w") as tenants_kustomization_file:
                yaml.safe_dump(
                    tenants_file,
                    tenants_kustomization_file,
                    sort_keys=False,
                    default_flow_style=False,
                )

            print("[green]Complete")
        else:
            print("[green]Skipped (tenant already listed)")

        # Move tenant file into env repo
        print("[blue]Move tenant flux .yaml file into local env repo...", end=" ")
        tenant_path_in_env_repo = os.path.join(env_tenant_folder, tenant_file_name)
        os.replace(tenant_path, tenant_path_in_env_repo)
        print("[green]Complete")

        # Sync changes (if any)
        print("[blue]Sync tenant file changes to repo (if any)...", end=" ")
        result = subprocess.run(
            ["git", "add", "."], cwd=env_repo_folder, capture_output=True, text=True
        )
        if result.returncode != 0:
            print(f"[red]Failed `git add` ({result.stderr})")
            return
        result = subprocess.run(
            ["git", "commit", "-m", f"Add tenant {tenant_name} to flux"],
            cwd=env_repo_folder,
            capture_output=True,
            text=True,
        )
        if "nothing to commit" in result.stdout.lower() and result.returncode != 0:
            nothing_to_commit = True
        elif result.returncode != 0:
            print(f"[red]Failed `git commit` ({result.stderr})")
            return
        else:
            nothing_to_commit = False

        if not nothing_to_commit:
            result = subprocess.run(
                ["git", "push", "origin", "master"],
                cwd=env_repo_folder,
                capture_output=True,
                text=True,
            )
            if result.returncode != 0:
                print(f"[red]Failed `git push` ({result.stderr})")
                return
        print("[green]Complete")

    # TOOD: Trigger reconciliation


def unregister_tenant_from_env(tenant_name):
    print("[blue]Reading values...", end=" ")
    tenant_keyvault_name = get_keyvault_name(tenant_name)
    env_name = get_keyvault_secret(tenant_keyvault_name, "env-name")
    env_keyvault_name = get_env_info(env_name).keyvault_name
    tenant_name = get_keyvault_secret(tenant_keyvault_name, "tenant-name")
    env_name = get_keyvault_secret(env_keyvault_name, "env-name")
    print("[green]Complete")

    # Delete tenant namespace and keyvault-sp-credentials secret
    print("[blue]Deleting tenant namespace and credentials secret...", end=" ")
    tenant_namespace = f"tenant-{tenant_name}"

    kubeconfig_content = get_env_kubectl(env_name)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as kc_file:
        kc_file.write(kubeconfig_content)
        kubeconfig_path = kc_file.name

    try:
        # Delete namespace (this also deletes all resources in it, including the secret)
        result = subprocess.run(
            [
                "kubectl",
                "--kubeconfig",
                kubeconfig_path,
                "delete",
                "namespace",
                tenant_namespace,
                "--ignore-not-found",
            ],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"[red]Failed to delete namespace ({result.stderr})")
            # Continue anyway to clean up flux files
        else:
            print("[green]Complete")
    finally:
        os.unlink(kubeconfig_path)

    # Remove tenant flux files from env repo
    with tempfile.TemporaryDirectory() as temp_folder:
        tenant_file_name = f"tenant-{tenant_name}.yaml"
        tenant_path = os.path.join(temp_folder, tenant_file_name)

        # Clone env repo
        print("[blue]Cloning env repo...", end=" ")
        env_repo_name = f"env-{env_name}"
        env_repo_url = f"https://github.com/btenvs/{env_repo_name}.git"
        env_repo_folder = os.path.join(temp_folder, "repo")

        result = subprocess.run(
            ["git", "clone", env_repo_url, env_repo_folder],
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            print("[green]Complete")
        else:
            print(f"[red]Failed ({result.stderr})")
            return

        # Remove tenant flux file from kustomization.yaml resource list
        print(
            "[blue]Delete tenant flux .yaml file from tenants kustomization list...",
            end=" ",
        )
        env_tenant_folder = os.path.join(env_repo_folder, "kubernetes", "tenants")
        tenant_kustomization_path = os.path.join(
            env_tenant_folder, "kustomization.yaml"
        )
        with open(tenant_kustomization_path, "r") as tenants_kustomization_file:
            tenants_file = yaml.safe_load(tenants_kustomization_file)

        if tenant_file_name in tenants_file["resources"]:
            tenants_file["resources"].remove(tenant_file_name)

            with open(tenant_kustomization_path, "w") as tenants_kustomization_file:
                yaml.safe_dump(
                    tenants_file,
                    tenants_kustomization_file,
                    sort_keys=False,
                    default_flow_style=False,
                )

            print("[green]Complete")
        else:
            print("[green]Skipped (tenant not found)")

        # Remove the tenant flux file
        print("[blue]Delete tenant flux .yaml file from local env repo...", end=" ")
        tenant_path_in_env_repo = os.path.join(env_tenant_folder, tenant_file_name)
        if os.path.exists(tenant_path_in_env_repo):
            os.remove(tenant_path_in_env_repo)
            print("[green]Complete")
        else:
            print("[green]Skipped (file not found)")

        # Sync changes (if any)
        print("[blue]Sync tenant file changes to repo (if any)...", end=" ")
        result = subprocess.run(
            ["git", "add", "."], cwd=env_repo_folder, capture_output=True, text=True
        )
        if result.returncode != 0:
            print(f"[red]Failed `git add` ({result.stderr})")
            return
        result = subprocess.run(
            ["git", "commit", "-m", f"Remove tenant {tenant_name} from flux"],
            cwd=env_repo_folder,
            capture_output=True,
            text=True,
        )
        if "nothing to commit" in result.stdout.lower() and result.returncode != 0:
            nothing_to_commit = True
        elif result.returncode != 0:
            print(f"[red]Failed `git commit` ({result.stderr})")
            return
        else:
            nothing_to_commit = False

        if not nothing_to_commit:
            result = subprocess.run(
                ["git", "push", "origin", "master"],
                cwd=env_repo_folder,
                capture_output=True,
                text=True,
            )
            if result.returncode != 0:
                print(f"[red]Failed `git push` ({result.stderr})")
                return
        print("[green]Complete")

    # Unregister tenant repo SSH key from env key vault
    tenant_key_name = env_tenant_ssh_key_name(tenant_name)
    delete_keyvault_secret(env_keyvault_name, tenant_key_name)

    return


def get_tenant_k8s_ns(tenant_name):
    return f"tenant-{tenant_name}"


def env_tenant_ssh_key_name(tenant_name):
    return f"tenant-repo-ssh-key-{tenant_name}"


def env_tenant_ssh_pub_key_name(tenant_name):
    return f"tenant-repo-ssh-pub-key-{tenant_name}"


def get_tenant_debug_telemetry_config(tenant_name, services=None):
    keyvault_name = get_keyvault_name(tenant_name)

    with tempfile.TemporaryDirectory() as temp_folder:
        kubeconfig_path = os.path.join(temp_folder, "kubeconfig")
        get_keyvault_secret_to_file(keyvault_name, "kubeconfig", kubeconfig_path)
        tenant_ns = get_tenant_k8s_ns(tenant_name)
        responses = parallel_pod_curl(
            kubeconfig_path, tenant_ns, "/debug/telemetry", services, port=8080
        )

        return responses


def set_tenant_debug_telemetry_loglevel(tenant_name, log_level, services=None):
    log_level_map = {
        "dbg": "debug",
        "info": "information",
        "warn": "warning",
        "err": "error",
    }
    dotnet_level = log_level_map.get(log_level.lower(), log_level.lower())
    keyvault_name = get_keyvault_name(tenant_name)

    with tempfile.TemporaryDirectory() as temp_folder:
        kubeconfig_path = os.path.join(temp_folder, "kubeconfig")
        get_keyvault_secret_to_file(keyvault_name, "kubeconfig", kubeconfig_path)
        tenant_ns = get_tenant_k8s_ns(tenant_name)
        responses = parallel_pod_curl(
            kubeconfig_path,
            tenant_ns,
            f"/debug/telemetry/loglevel?level={dotnet_level}",
            services,
            method="POST",
            port=8080,
        )

        return responses


def set_tenant_debug_telemetry_tracing(tenant_name, tracing_enabled, services=None):
    keyvault_name = get_keyvault_name(tenant_name)

    with tempfile.TemporaryDirectory() as temp_folder:
        kubeconfig_path = os.path.join(temp_folder, "kubeconfig")
        get_keyvault_secret_to_file(keyvault_name, "kubeconfig", kubeconfig_path)
        tenant_ns = get_tenant_k8s_ns(tenant_name)
        responses = parallel_pod_curl(
            kubeconfig_path,
            tenant_ns,
            f"/debug/telemetry/tracing?enabled={str(tracing_enabled).lower()}",
            services,
            method="POST",
            port=8080,
        )

        return responses


def listen_to_tenant_eventhub(
    tenant, eh_name, consumer_group="monitor", show_full_event=True
):
    keyvault_name = get_keyvault_name(tenant)
    secret_name_cs_r = f"{eh_name}-eh-connection-string-r"
    secret_name_cs_rw = f"{eh_name}-eh-connection-string-rw"
    eh_connection_string = get_keyvault_secret(
        keyvault_name, secret_name_cs_r
    ) or get_keyvault_secret(keyvault_name, secret_name_cs_rw)

    if not eh_connection_string:
        print(
            f"[red]Tenant event hub connection string not found for event hub {eh_name}"
        )
        return

    def on_event(partition_context, event):
        prefix = f"[blue][{partition_context.partition_id}][/blue]"
        if show_full_event:
            print(f"{prefix} {event}")
        else:
            print(f"{prefix} {event.body_as_str()}")

    client = EventHubConsumerClient.from_connection_string(
        eh_connection_string, consumer_group
    )
    print(f"[blue]Listening to {client.eventhub_name}...")
    try:
        with client:
            client.receive(on_event=on_event, starting_position="@latest")
    except KeyboardInterrupt:
        print("[blue]Stopped listening")  # Doesn't work


@dataclass
class TenantLoginInfo:
    email: str
    password: str
    url: str


def get_tenant_login_info(tenant) -> TenantLoginInfo:
    keyvault_name = get_keyvault_name(tenant)
    email = get_keyvault_secret(keyvault_name, "user-service-default-user-email")
    password = get_keyvault_secret(keyvault_name, "user-service-default-user-password")
    tenant_domain = get_keyvault_secret(keyvault_name, "tenant-domain")
    url = f"https://{tenant_domain}"

    return TenantLoginInfo(email, password, url)



import sys
from pathlib import Path

import chevron
from rich import print


def to_absolute_path(current_file, relative_path: str) -> Path:
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        # noinspection PyProtectedMember
        script_dir = Path(sys._MEIPASS)
    else:
        script_dir = Path(current_file).parent

    return script_dir / relative_path


def generate_file_from_template(template_filename, out_filename, data):
    print(f"[blue]Generating file {out_filename}...", end=" ")
    with open(template_filename, "r") as template_file:
        generated_content = chevron.render(template_file, data)

    with open(out_filename, "w") as file:
        file.write(generated_content)

    print("[green]Complete")



import json
import os
import shutil
import stat
import subprocess
import time
from dataclasses import asdict, dataclass

from rich import print

from btpy.configuration.config import KeyVaultInfo, get_config_value
from btpy.core.azure_utility import (
    assign_subscription_role,
    create_service_principal,
    delete_service_principal,
    get_current_subscription,
    get_keyvault_secret,
    get_keyvault_secret_to_file,
    get_service_principals_by_name,
    get_storage_table_entry,
    list_storage_table_entries,
    location_tag,
    set_keyvault_admin_rbac,
    upload_keyvault_secret,
    upload_keyvault_secret_file,
    upsert_table_storage_entry,
)
from btpy.core.file_utility import generate_file_from_template, to_absolute_path
from btpy.core.misc_utility import (
    create_domain,
    get_gh_btenvs_app_info,
    init_git_repo,
    setup_ssh_keys,
)
from btpy.core.rancher_utility import (
    get_or_create_rancher_azure_credential,
    get_rancher_mgmt_info,
)
from btpy.core.iac_utility import bootstrap_tf_repo, get_keyvault_name
from btpy.iac.suffix import generate

ENV_LIST_TABLE_NAME = "envs"


@dataclass
class EnvListInfo:
    tenant_id: str
    subscription_id: str
    name: str


@dataclass
class EnvInfo:
    tenant_id: str
    subscription_id: str
    name: str
    keyvault_name: str


@dataclass
class EnvTemplateData:
    customer: str
    environment: str
    resource_group: str
    location: str
    shloc: str
    suffix: str
    keyvault_name: str
    tenant_id: str
    sp_client_id: str
    sp_client_secret: str
    rancher_api_url: str
    rancher_api_token: str
    rancher_cloud_creds_name: str
    gh_btenvs_app_id: str
    gh_btenvs_app_private_key_path: str
    gh_btenvs_installation_id: str
    env_domain: str


def list_envs() -> list[EnvInfo]:
    env_list_info = _get_env_list_info()
    query_result = list_storage_table_entries(
        env_list_info.name, ENV_LIST_TABLE_NAME, env_list_info.subscription_id
    )
    return [EnvInfo(**json.loads(entry["Data"])) for entry in query_result["items"]]


def get_env_info(env_name) -> EnvInfo | None:
    env_list_info = _get_env_list_info()
    query_result = get_storage_table_entry(
        env_list_info.name,
        ENV_LIST_TABLE_NAME,
        env_list_info.subscription_id,
        env_name,
        env_name,
    )

    if query_result:
        return EnvInfo(**json.loads(query_result["Data"]))
    return None


@dataclass
class EnvCreationOptions:
    customer: str
    environment: str
    location: str
    suffix: str | None = None
    regen_ssh: bool = False
    regen_sp: bool = False
    subdomain: str | None = None


def create_env(options: EnvCreationOptions):
    if options.suffix is None:
        options.suffix = generate()["suffix"]

    shloc = location_tag(options.location)

    env_name = f"{options.customer}-{options.environment}-{shloc}{options.suffix}"
    folder = f"env-{options.customer}-{options.environment}-{shloc}{options.suffix}"
    infra_folder = os.path.join(folder, "terraform")
    flux_infra_folder = os.path.join(folder, "kubernetes", "infrastructure", "configs")

    print(f"[cyan]Creating env {env_name}")

    # Clone env template repo
    if not os.path.isdir(folder):

        def __on_rm_error(func, path, exc):
            os.chmod(path, stat.S_IWRITE)
            func(path)

        repo_url = "https://github.com/beacontower/env-template.git"
        subprocess.run(
            ["git", "clone", "--branch", "master", repo_url, folder], check=True
        )
        shutil.rmtree(os.path.join(folder, ".git"), onerror=__on_rm_error)
    else:
        print(f"[blue]Repo folder {folder} already exists")

    # Set up resource group, storage account for TF and key vault
    tf_data = bootstrap_tf_repo(
        options.customer,
        options.environment,
        options.location,
        options.suffix,
        prefix="env",
        folder=infra_folder,
    )
    time.sleep(6)  # Wait and hope role assignment has refreshed cache on the servers

    if tf_data is None:
        return

    keyvault_name = tf_data.keyvault_name

    # Register env in env list
    _add_env_to_storage(env_name, keyvault_name)

    # Generate and store git repo SSH key
    secrets_folder_name = ".secrets"
    secrets_path = os.path.join(folder, "terraform", secrets_folder_name)
    if not os.path.isdir(secrets_path):
        os.mkdir(secrets_path)

    setup_ssh_keys(
        keyvault_name,
        secrets_path,
        "Auto-generated BT env repo SSH keys",
        options.regen_ssh,
    )

    # Set up service principal used for key vault and monitoring access
    sp = _setup_service_principal(
        f"sp-bt-env-{env_name}",
        keyvault_name,
        tf_data.keyvault_id,
        tf_data.subscription_id,
    )

    # Misc secrets
    upload_keyvault_secret(keyvault_name, "env-name", env_name)
    rancher_credentials = get_or_create_rancher_azure_credential(
        tf_data.tenant_id, tf_data.subscription_id
    )
    _setup_charts_ssh_key(keyvault_name, secrets_path)
    upload_keyvault_secret(
        keyvault_name,
        "otel-url",
        "http://otel-collector.monitoring.svc.cluster.local:4317",
    )
    upload_keyvault_secret(keyvault_name, "az-tenant-id", tf_data.tenant_id)
    upload_keyvault_secret(keyvault_name, "az-subscription-id", tf_data.subscription_id)

    # Get github btenvs repo auth info and store in secret file
    gh_btenvs_app_info = get_gh_btenvs_app_info()
    gh_btenvs_app_key_filename = "gh_betenvs_app_key.pem"
    gh_btenvs_app_private_key_path = os.path.join(
        secrets_path, gh_btenvs_app_key_filename
    )
    relative_gh_btenvs_app_private_key_path = os.path.join(
        secrets_folder_name, gh_btenvs_app_key_filename
    )
    with open(gh_btenvs_app_private_key_path, "w") as key_file:
        key_file.write(gh_btenvs_app_info.private_key)

    rancher_mgmt_info = get_rancher_mgmt_info()

    env_domain = create_domain(options.subdomain)
    upload_keyvault_secret(keyvault_name, "env-domain", env_domain)

    env_data = EnvTemplateData(
        customer=tf_data.customer,
        environment=tf_data.environment,
        resource_group=tf_data.resource_group,
        location=tf_data.location,
        shloc=tf_data.shloc,
        suffix=tf_data.suffix,
        keyvault_name=keyvault_name,
        tenant_id=tf_data.tenant_id,
        sp_client_id=sp["appId"],
        sp_client_secret=sp["password"],
        rancher_api_url=rancher_mgmt_info.url,
        rancher_api_token=rancher_mgmt_info.token,
        rancher_cloud_creds_name=rancher_credentials.name,
        gh_btenvs_app_id=gh_btenvs_app_info.app_id,
        gh_btenvs_app_private_key_path=relative_gh_btenvs_app_private_key_path,
        gh_btenvs_installation_id=gh_btenvs_app_info.installation_id,
        env_domain=env_domain,
    )
    generate_file_from_template(
        to_absolute_path(__file__, "templates/env-terraform.tfvars.json.mustache"),
        os.path.join(infra_folder, "terraform.tfvars.json"),
        asdict(env_data),
    )
    generate_file_from_template(
        to_absolute_path(__file__, "templates/env-eso.yaml.mustache"),
        os.path.join(flux_infra_folder, "external-secrets.yaml"),
        asdict(env_data),
    )
    generate_file_from_template(
        to_absolute_path(__file__, "templates/env-config.yaml.mustache"),
        os.path.join(flux_infra_folder, "env-config.yaml"),
        asdict(env_data),
    )

    # Prepare git repo
    if not init_git_repo(folder, "env"):
        return


def _setup_charts_ssh_key(env_keyvault, secret_folder_path):
    # Copy from global key vault
    global_kv_info = KeyVaultInfo(**get_config_value("global_keyvault"))
    key_path = os.path.join(secret_folder_path, "charts_repo_ssh_key")
    get_keyvault_secret_to_file(global_kv_info.name, "charts-repo-ssh-key", key_path)
    upload_keyvault_secret_file(env_keyvault, "charts-repo-ssh-key", key_path)


def _setup_service_principal(name, keyvault_name, keyvault_id, subscription_id):
    existing_sp_client_id = get_keyvault_secret(keyvault_name, "sp-client-id")
    existing_sp_client_secret = get_keyvault_secret(keyvault_name, "sp-client-secret")

    if existing_sp_client_id and existing_sp_client_secret:
        # Ensure existing SP has Contributor role
        set_keyvault_admin_rbac(keyvault_id, existing_sp_client_id)
        _setup_service_principal_roles(subscription_id, existing_sp_client_id)
        return {"appId": existing_sp_client_id, "password": existing_sp_client_secret}

    # Clean up potentially existing SPs
    existing_sps = get_service_principals_by_name(name)

    if existing_sps:
        for sp in existing_sps:
            delete_service_principal(sp["id"])

    sp = create_service_principal(name)
    set_keyvault_admin_rbac(keyvault_id, sp["appId"])

    # Assign Contributor role for Azure resource management (e.g., disk provisioning)
    _setup_service_principal_roles(subscription_id, sp["appId"])

    upload_keyvault_secret(keyvault_name, "sp-client-id", sp["appId"])
    upload_keyvault_secret(keyvault_name, "sp-client-secret", sp["password"])

    return sp


def _setup_service_principal_roles(subscription_id, sp_client_id):
    assign_subscription_role(subscription_id, sp_client_id, "Contributor")
    assign_subscription_role(subscription_id, sp_client_id, "Monitoring Reader")


def _add_env_to_storage(env_name, keyvault_name):
    print("[blue]Adding env to global env list...", end=" ")
    env_list_info = _get_env_list_info()
    list_sub_id = env_list_info.subscription_id
    list_name = env_list_info.name
    sub_info = get_current_subscription()
    env_info = EnvInfo(
        tenant_id=sub_info["tenantId"],
        subscription_id=sub_info["id"],
        name=env_name,
        keyvault_name=keyvault_name,
    )
    min_env_data = json.dumps(asdict(env_info), separators=(",", ":"))
    upsert_table_storage_entry(
        list_name,
        ENV_LIST_TABLE_NAME,
        env_name,
        env_name,
        min_env_data,
        list_sub_id,
        log=False,
    )
    print("[green]Complete")


def get_env_kubectl(env_name):
    env_info = get_env_info(env_name)
    return get_keyvault_secret(env_info.keyvault_name, "kubeconfig")


def _get_env_list_info() -> EnvListInfo:
    env_acc_config = KeyVaultInfo(**get_config_value("env_list_storage_account"))
    return EnvListInfo(
        tenant_id=env_acc_config.tenant,
        subscription_id=env_acc_config.subscription,
        name=env_acc_config.name,
    )


@dataclass
class EnvMonitorInfo:
    url: str


def get_env_monitor_info(env) -> EnvMonitorInfo:
    keyvault_name = get_keyvault_name(env)
    env_domain = get_keyvault_secret(keyvault_name, "env-domain")

    return EnvMonitorInfo(f"https://{env_domain}/monitor")



import os
from dataclasses import asdict, dataclass

from rich import print

from btpy.core.azure_utility import (
    create_container,
    create_keyvault,
    create_resource_group,
    create_storage_account,
    get_azure_tenant_id,
    get_current_subscription,
    get_storage_connection_string,
    location_tag,
)
from btpy.core.file_utility import generate_file_from_template, to_absolute_path
from btpy.iac.suffix import generate


@dataclass
class TerraformBootstrapData:
    resource_group: str
    storage_account_name: str
    container_name: str
    tf_blob_name: str
    keyvault_name: str
    keyvault_id: str
    customer: str
    environment: str
    env_prefix: str | None
    location: str
    shloc: str
    suffix: str
    tenant_id: str
    subscription_id: str


# TODO: Move out from this file? Rename file?
def bootstrap_tf_repo(customer, env, location, suffix=None, prefix=None, folder=None):
    if suffix is None:
        suffix = generate()["suffix"]

    shloc = location_tag(location)
    repo_name = f"{customer}-{env}-{shloc}{suffix}"
    repo_name_no_hyphens = get_no_hyphen_name(repo_name)
    customer_env_name = f"{customer}-{env}"
    customer_env_name_no_hyphens = f"{customer}{env}"
    keyvault_name = get_keyvault_name(repo_name)

    if len(keyvault_name) > 24:
        print(f"[red]Key vault name {keyvault_name} is too long (24 is max)")
        raise f"Key vault name {keyvault_name} is too long (24 is max)"

    if prefix is not None:
        customer_env_name = f"{prefix}-{customer_env_name}"

    resource_group_name = f"rg-bt-{repo_name}"
    storage_account_name = f"stgtf{repo_name_no_hyphens}"
    storage_account_container = "tfstate"
    storage_account_blob_name = "terraform.tfstate"
    tenant_id = get_azure_tenant_id()

    try:
        create_resource_group(resource_group_name, location)
        create_storage_account(resource_group_name, storage_account_name, location)
        storage_connection_string = get_storage_connection_string(storage_account_name)
        create_container(storage_connection_string, storage_account_container)
        keyvault_info = create_keyvault(resource_group_name, keyvault_name, location)
    except Exception as e:
        print(f"[red]Terraform bootstrapping failed ({e})")
        return None

    subscription_id = get_current_subscription()["id"]
    data = TerraformBootstrapData(
        resource_group=resource_group_name,
        storage_account_name=storage_account_name,
        container_name=storage_account_container,
        tf_blob_name=storage_account_blob_name,
        keyvault_name=keyvault_name,
        keyvault_id=keyvault_info["id"],
        customer=customer,
        environment=env,
        env_prefix=prefix,
        location=location,
        shloc=shloc,
        suffix=suffix,
        tenant_id=tenant_id,
        subscription_id=subscription_id,
    )

    backend_path = os.path.join(folder if folder is not None else "", "backend.tf")
    generate_file_from_template(
        to_absolute_path(__file__, "../iac/templates/backend.mustache"),
        backend_path,
        asdict(data),
    )

    return data


def get_keyvault_name(name):
    return f"kvbt{get_no_hyphen_name(name)}"


def get_no_hyphen_name(name):
    return name.replace("-", "")



import json
import os
import secrets
import string
import subprocess
from dataclasses import dataclass

import git
from rich import print

from btpy.configuration.config import KeyVaultInfo, get_config_value, get_config
from btpy.core.azure_utility import (
    assign_subscription_role,
    create_service_principal,
    get_keyvault_secret,
    query_keyvault_secrets,
    upload_keyvault_secret,
    upload_keyvault_secret_file,
)


@dataclass
class AzureServicePrincipal:
    client_id: str
    client_secret: str


def generate_keys_ed25519(private_key_path, public_key_path, comment):
    if os.path.exists(private_key_path) or os.path.exists(public_key_path):
        print("[blue]Key file(s) already exist")
        return

    subprocess.run(
        [
            "ssh-keygen",
            "-t",
            "ed25519",
            "-C",
            comment,
            "-f",
            private_key_path,
            "-N",
            "",
        ],
        check=True,
    )
    os.chmod(private_key_path, 0o600)
    os.chmod(public_key_path, 0o644)


def setup_ssh_keys(keyvault_name, secrets_path, comment, regen_ssh=False):
    ssh_path = os.path.join(secrets_path, "id_ed25519")
    ssh_pub_path = os.path.join(secrets_path, "id_ed25519.pub")

    existing_ssh_key = get_keyvault_secret(keyvault_name, "repo-ssh-key")
    existing_ssh_pub_key = get_keyvault_secret(keyvault_name, "repo-ssh-key-pub")

    if existing_ssh_key and existing_ssh_pub_key and not regen_ssh:
        return

    generate_keys_ed25519(ssh_path, ssh_pub_path, comment)
    upload_keyvault_secret_file(keyvault_name, "repo-ssh-key", ssh_path)
    upload_keyvault_secret_file(keyvault_name, "repo-ssh-key-pub", ssh_pub_path)
    os.remove(ssh_path)
    os.remove(ssh_pub_path)


def init_git_repo(folder, repo_kind_message):
    print(f"[blue]Set up local {repo_kind_message} git repo...", end=" ")

    # Initialize or open existing repo
    try:
        repo = git.Repo.init(folder)
    except git.exc.GitCommandError as e:
        if "already exists" not in str(e):
            print(f"[red]Failed (git init - {e})")
            return False
        repo = git.Repo(folder)

    # If the repo is already initialized, fetch and pull before committing
    try:
        if repo.remotes:
            origin = repo.remotes.origin
            origin.fetch()
            origin.pull()
    except git.exc.GitCommandError as e:
        if "There is no tracking information for the current branch" not in str(e):
            print(f"[red]Failed (git fetch/pull - {e})")
            return False

    # Stage and commit changes
    try:
        # Check if there are changes to commit BEFORE staging
        if not repo.head.is_valid():
            # No commits yet - check for any files
            has_changes = len(repo.untracked_files) > 0
        else:
            # Has commits - check for changes
            has_changes = (
                len(repo.index.diff(None)) > 0  # Working tree changes
                or len(repo.index.diff("HEAD")) > 0  # Staged changes
                or len(repo.untracked_files) > 0  # Untracked files
            )

        if has_changes:
            repo.git.add(
                "."
            )  # Use git.add() to properly respect .gitignore and exclude .git/
            repo.index.commit(
                f"{repo_kind_message.capitalize()} repo bootstrap commit by btpy"
            )
        # else: nothing to commit, silently skip (matches original behavior)

    except git.exc.GitCommandError as e:
        print(f"[red]Failed (git commit - {e})")
        return False

    print("[green]Complete")
    return True


def get_azure_tenant_sps():
    global_kv_info = KeyVaultInfo(**get_config_value("global_keyvault"))

    if not global_kv_info:
        print("[red]Failed to read config for global keyvault")
        return []

    sps = query_keyvault_secrets(
        global_kv_info.name, "[?starts_with(name,'az-tenant-sp')]"
    )
    return [sp["name"] for sp in sps]


def get_or_create_azure_tenant_sp(
    tenant_id, subscription_id=None
) -> AzureServicePrincipal:
    kv_info = KeyVaultInfo(**get_config_value("global_keyvault"))
    kv_name = kv_info.name
    sps = get_azure_tenant_sps()
    sp_name = next((sp for sp in sps if tenant_id in sp), None)

    if not sp_name:
        sp = create_azure_tenant_sp(kv_name, tenant_id, subscription_id)["value"]
    else:
        sp = get_keyvault_secret(kv_name, sp_name)

    sp_obj = AzureServicePrincipal(**json.loads(sp))

    if subscription_id:
        assign_subscription_role(subscription_id, sp_obj.client_id, "Contributor")

    return sp_obj


def create_azure_tenant_sp(keyvault_name, tenant_id, subscription_id=None):
    sp_name = f"az-tenant-sp-{tenant_id}"
    sp = create_service_principal(sp_name, tenant_id)

    sp_secret = {"client_id": sp["appId"], "client_secret": sp["password"]}
    return upload_keyvault_secret(
        keyvault_name, sp_name, json.dumps(sp_secret, separators=(",", ":"))
    )


@dataclass
class GitHubAuthInfo:
    private_key: str
    app_id: str
    installation_id: str


def get_gh_btenvs_app_info():
    kv_info = KeyVaultInfo(**get_config_value("global_keyvault"))
    return GitHubAuthInfo(
        private_key=get_keyvault_secret(kv_info.name, "btenvs-private-key"),
        app_id="2251489",  # App ID taken from GitHub website on the app's page
        installation_id="93545915",  # Installation ID taken from the URL in the btenvs repo -> settings -> github apps -> configure the installed app
    )


def create_domain(subdomain):
    btpy_config = get_config()
    base_domain = btpy_config.domain
    return f"{subdomain}.{base_domain}"


def generate_password(
    length: int = 20,
    lowercase: bool = True,
    uppercase: bool = True,
    digits: bool = True,
    symbols: str | None = "_",
) -> str:
    """
    Generate a secure password with guaranteed character type representation.

    Args:
        length: Total password length (must be >= number of enabled types)
        lowercase: Include lowercase letters (a-z)
        uppercase: Include uppercase letters (A-Z)
        digits: Include digits (0-9)
        symbols: String of allowed symbols, or None to disable

    Returns:
        Generated password with at least one of each enabled type
    """
    pools = []
    if lowercase:
        pools.append(string.ascii_lowercase)
    if uppercase:
        pools.append(string.ascii_uppercase)
    if digits:
        pools.append(string.digits)
    if symbols:
        pools.append(symbols)

    if not pools:
        raise ValueError("At least one character type must be enabled")

    if length < len(pools):
        raise ValueError(
            f"Length must be at least {len(pools)} to include one of each type"
        )

    # Guarantee at least one from each pool
    password_chars = [secrets.choice(pool) for pool in pools]

    # Fill remaining length from combined pool
    all_chars = "".join(pools)
    password_chars.extend(secrets.choice(all_chars) for _ in range(length - len(pools)))

    # Shuffle to avoid predictable positions
    secrets.SystemRandom().shuffle(password_chars)

    return "".join(password_chars)



import asyncio
from types import SimpleNamespace

from btpy.core.envs.env_desc import EnvDesc
from btpy.core.envs.env_utility import to_resource_name
from btpy.core.resource_clients.resource_client import ResourceStatus
from btpy.core.resource_clients.resource_client_factory import get_resource_client


class EnvClient:
    def __init__(self, env_desc: EnvDesc):
        self.desc = env_desc

    async def status(self):
        resource_clients = self.get_resource_clients()
        named_tasks = {
            resource_client.name: asyncio.create_task(resource_client.client.status())
            for resource_client in resource_clients
        }
        status_results = await asyncio.gather(*named_tasks.values())

        service_statuses = [
            SimpleNamespace(name=name, status=result)
            for (name, result) in zip(named_tasks.keys(), status_results)
        ]
        env_status = self._get_env_status_from_service_statuses(service_statuses)

        [await resource_client.client.close() for resource_client in resource_clients]

        return SimpleNamespace(status=env_status, services=service_statuses)

    def get_resource_client(self, resource_name):
        resource_name_parts = resource_name.split(".")
        service_name = resource_name_parts[0]
        resource_name = resource_name_parts[1]
        service_desc = next(
            (service for service in self.desc.services if service.name == service_name),
            None,
        )

        if service_desc is None:
            return None

        resource_desc = next(
            (
                resource
                for resource in service_desc.resources
                if resource.name == resource_name
            ),
            None,
        )

        if resource_desc is None:
            return None

        return get_resource_client(resource_desc)

    def get_resource_clients(self):
        resource_clients = []

        for service in self.desc.services:
            for resource in service.resources:
                client = get_resource_client(resource)
                resource_name = to_resource_name(service.name, resource.name)
                resource_clients.append(
                    SimpleNamespace(name=resource_name, client=client)
                )

        return resource_clients

    @staticmethod
    def _get_env_status_from_service_statuses(service_statuses):
        if all(
            service.status == service_statuses[0].status for service in service_statuses
        ):
            return service_statuses[0].status

        return ResourceStatus.Mixed



import json
import os
import random
import string

from btpy import application, environment, misc

logger = misc.setup_logging()


def generate(prefix=""):
    suffix = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(4)
    )
    return {f"{prefix}suffix": suffix}


def write(data, filename="suffix.json", rewrite=False):
    """
    Write suffix to file. If file exists, ask user if they want to overwrite.
    return if we allow re-writing of suffixes.
    """
    try:
        if rewrite:
            misc.write_json(filename, data)
            return True
        existing_suffix = get_data(filename, in_parent_directories=False)
        key = os.path.basename(filename).split(".")[0]
        new_suffix = input(
            f"Suffix already exists. ({existing_suffix[key]}) Do you want to generate a new one? (N/y): "
        )
        if new_suffix == "y":
            misc.write_json(filename, data)
            return True
        else:
            return False
    except FileNotFoundError:
        logger.info("Suffix file not found. Generating a new one.")
        misc.write_json(filename, data)


def get_data(filename="suffix.json", in_parent_directories=True):
    path = (
        misc.find_file_in_parent_directories(filename)
        if in_parent_directories
        else filename
    )
    with open(path, "r") as f:
        return json.load(f)


def write_global_suffix(suffix=None, rewrite=False):
    data = (
        generate(prefix="global_")
        if not suffix
        else {"global_suffix": suffix["suffix"]}
    )
    app_dir = application.get_directory()
    write(data, filename=os.path.join(app_dir, "global_suffix.json"), rewrite=rewrite)


def get_global_suffix():
    app_dir = application.get_directory()
    return get_data(filename=os.path.join(app_dir, "global_suffix.json"))


def write_environment_suffix(suffix=None, rewrite=False):
    data = (
        generate(prefix="environment_")
        if not suffix
        else {"environment_suffix": suffix["suffix"]}
    )
    env_dir = environment.get_directory()
    write(
        data, filename=os.path.join(env_dir, "environment_suffix.json"), rewrite=rewrite
    )


def get_environment_suffix():
    env_dir = environment.get_directory()
    return get_data(filename=os.path.join(env_dir, "environment_suffix.json"))



import json
import os

from btpy import variables


def create_tfvars_directory(directory):
    """
    Create a tfvars directory in a directory.
    """
    tfvars_dir = os.path.join(directory, "tfvars")
    if not os.path.exists(tfvars_dir):
        os.mkdir(tfvars_dir)
    return tfvars_dir


def write_config_data_to_tfvars_json_file(tfname, tfvars_dir, data):
    """
    Write a <tfname>.tfvars.json file to a tfvar directory.
    """
    # reformat data to be compatible with terraform output format
    output = {}
    for key, value in data.items():
        output[key] = {"value": value}
    write_tfvars_json_file(tfname, tfvars_dir, output)


def write_tfvars_json_file(tfname, tfvars_dir, tfvars_json):
    """
    Write a <tfname>.tfvars.json file to a tfvar directory.
    """
    with open(os.path.join(tfvars_dir, f"{tfname}.tfvars.json"), "w") as f:
        f.write(json.dumps(tfvars_json, indent=4))


def find_all_directories(starting_directory, excluded_dirs=[]):
    """
    find all tfvars directories in all subdirectories of a starting directory.
    This can be used to find all directories where a to update a tfvars.json file.

    exclude is a list of directories to exclude from the returned list.

    """
    for root, dirs, files in os.walk(starting_directory):
        if "tfvars" in dirs:
            if root not in excluded_dirs:
                yield os.path.join(root, "tfvars")


def merge_all_tfvars(tfvars_dir):
    """
    Merge all tfvars files in the tfvars directory into a single dictionary.
    """
    # get all tfvars files in the tfvars directory
    tfvars_files = [
        os.path.join(tfvars_dir, file)
        for file in sorted(os.listdir(tfvars_dir))
        if file.endswith(".tfvars.json")
    ]

    # merge all tfvars json files
    merged_tfvars = {}
    for tfvars_file in tfvars_files:
        try:
            with open(tfvars_file, "r") as f:
                merged_tfvars = {**merged_tfvars, **json.load(f)}
        except ValueError:
            print(f"Error reading {tfvars_file}")

    return merged_tfvars


def write_file_and_update(output, tfvars_dir, tfname):
    """
    Write <output> to <tfname>.tfvars.json file in <tfvars_dir> and run the
    create_tfvars.py to generate a new terraform.tvars.json in parent dir of <tfvars_dir>.
    """
    write_tfvars_json_file(tfname, tfvars_dir, output)
    merged_tfvars = merge_all_tfvars(tfvars_dir)
    drive, path = os.path.splitdrive(tfvars_dir)
    tfvars_dir_list = path.split(os.sep)
    parent_dir = os.path.join(os.sep, *tfvars_dir_list[:-1])
    if drive:
        parent_dir = os.path.join("C:\\", parent_dir)

    vars = variables.parse_tf_file(os.path.join(parent_dir, "variables.tf"))

    create_filtered_tfvars_file(parent_dir, vars, merged_tfvars)


def create_filtered_tfvars_file(dir, variables, merged_tfvars):
    """
    Create a tfvars json file with only the merged_tfvars entries present as variables in variable.tf
    """
    filtered_variables = [
        variable for variable in variables if variable in merged_tfvars.keys()
    ]
    with open(os.path.join(dir, "terraform.tfvars.json"), "w") as f:
        json.dump(
            {
                variable: merged_tfvars[variable]["value"]
                for variable in filtered_variables
            },
            f,
            indent=4,
        )



import json
import shlex
import subprocess


def kubectl(kubeconfig_path, command):
    if isinstance(command, list):
        args = ["kubectl", f"--kubeconfig={kubeconfig_path}"] + command
    else:
        args = ["kubectl", f"--kubeconfig={kubeconfig_path}", *shlex.split(command)]

    # print(args)
    program = subprocess.run(args, capture_output=True, text=True)

    stdout_data = None
    # print(program.stdout)
    # print("-------")
    # print(program.stderr)
    if program.stdout and program.stdout.strip():
        try:
            stdout_data = json.loads(program.stdout)
        except json.JSONDecodeError:
            stdout_data = program.stdout

    return program.returncode, stdout_data, program.stderr


def parallel_pod_curl(
    kubeconfig_path,
    namespace,
    endpoint,
    services=None,
    method="GET",
    body=None,
    port=80,
):
    """
    DO NOT TOUCH THIS FUNCTION. The escaping works. Do not ask why.

    Returns result as a set of lines of the format `<pod_name>|<curl_response>`
    """
    if services:
        services_str = ",".join(services)
        label_selector = f"app.kubernetes.io/name in ({services_str})"
    else:
        label_selector = "beacontower.io/telemetry=true"
    output_format = r'{range .items[*]}{.metadata.name}{"\t"}{.status.podIP}{"\n"}{end}'
    _, output, _ = kubectl(
        kubeconfig_path,
        f"get pods -n {namespace} -l '{label_selector}' -o jsonpath='{output_format}'",
    )

    if not output:
        return {}

    pods = [line.split("\t") for line in output.strip().split("\n") if line]
    ip_to_pod = {pod[1]: pod[0] for pod in pods}
    pod_ips = [pod[1] for pod in pods]
    ips_str = " ".join(pod_ips)
    optional_port_part = f":{port}" if port else ""
    if method == "GET":
        curl_cmd = f"curl -s http://{{}}{optional_port_part}{endpoint}"
    else:
        curl_cmd = f"curl -s -X {method} http://{{}}{optional_port_part}{endpoint}"
        if body:
            curl_cmd += f" -H 'Content-Type: application/json' -d '{body}'"

    inner_cmd = f"printf '%s\\n' {ips_str} | xargs -P 20 -I {{}} sh -c 'echo \"{{}}|$({curl_cmd})\"'"
    cmd = ["exec", "-n", namespace, "deploy/nats-box", "--", "sh", "-c", inner_cmd]

    _, output, stderr = kubectl(kubeconfig_path, cmd)

    responses = {}
    for line in output.strip().split("\n"):
        if "|" in line:
            line_parts = line.split("|", 1)
            ip = line_parts[0].strip()
            json_str = line_parts[1].strip()
            if not ip or not json_str:
                continue
            pod_name = ip_to_pod.get(ip, ip)
            responses[pod_name] = json.loads(json_str)

    return responses



import asyncio
from typing import Annotated

from cyclopts import App, Parameter
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from btpy.configuration.aliases import (
    load_aliases,
    remove_alias,
    resolve_env_alias,
    set_alias,
)
from btpy.configuration.config import get_config_path
from btpy.core.envs import env_desc_loader
from btpy.core.envs.env2 import (
    EnvCreationOptions,
    create_env,
    get_env_kubectl,
    list_envs,
    get_env_monitor_info,
)
from btpy.core.envs.env_client import EnvClient
from btpy.core.envs.env_desc_loader import load_env_version
from btpy.core.envs.env_migrator import migrate_env
from btpy.print_utility import print_status

app = App()


@app.command(name="create")
async def create_env_cmd(
    *,
    customer: Annotated[str, Parameter(name=["-c", "--customer"], required=True)],
    environment: Annotated[str, Parameter(name=["-e", "--environment"], required=True)],
    location: Annotated[str, Parameter(name=["-l", "--location"], required=True)],
    suffix: Annotated[str, Parameter(name=["-s", "--suffix"])] = None,
    regen_ssh: Annotated[bool, Parameter(name=["--regen-ssh"])] = False,
    regen_sp: Annotated[bool, Parameter(name=["--regen-sp"])] = False,
):
    """Create a new environment with specified parameters.

    Args:
        -c, --customer: Customer name or identifier
        -e, --environment: Environment name (e.g., dev, test, prod)
        -l, --location: Azure region location
        -s, --suffix: Optional 4 char custom suffix for environment name, will be randomized otherwise
        --regen-ssh: If True, regenerate SSH keys even if they exist
        --regen-sp: If True, regenerate service principal credentials even if they exist
    """
    options = EnvCreationOptions(
        customer=customer,
        environment=environment,
        location=location,
        suffix=suffix,
        regen_ssh=regen_ssh,
        regen_sp=regen_sp,
    )
    create_env(options)


@app.command(name="list")
async def list_envs_cmd():
    env_list = list_envs()
    table = Table()

    table.add_column("env")
    table.add_column("desc")
    table.add_column("keyvault")
    table.add_column("sub")
    table.add_column("tenant")

    for env in env_list:
        table.add_row(
            env.name, "", env.keyvault_name, env.subscription_id, env.tenant_id
        )

    print(table)


alias_app = App(name="alias", help="Manage environment name aliases")


@alias_app.command(name="set")
def alias_set_cmd(alias: str, env_name: str):
    """Set an alias for an environment name.

    Args:
        alias: Short name to use (e.g., 'test')
        env_name: Full environment name (e.g., 'glaze-test-neu1234')
    """
    set_alias(alias, env_name)
    print(f"[green]Alias '{alias}' -> '{env_name}' set successfully")


@alias_app.command(name="list")
def alias_list_cmd():
    """List all configured aliases."""
    aliases = load_aliases()

    if not aliases:
        print("[yellow]No aliases configured")
        return

    table = Table(title="Environment Aliases")
    table.add_column("Alias", style="cyan")
    table.add_column("Environment Name", style="blue")

    for alias, env_name in sorted(aliases.items()):
        table.add_row(alias, env_name)

    print(table)


@alias_app.command(name="remove")
def alias_remove_cmd(alias: str):
    """Remove an alias.

    Args:
        alias: The alias to remove
    """
    if remove_alias(alias):
        print(f"[green]Alias '{alias}' removed successfully")
    else:
        print(f"[yellow]Alias '{alias}' not found")


app.command(alias_app)


@app.command(name="connect")
async def connect_env_cmd(env_name: str):
    env_name = resolve_env_alias(env_name)
    kubeconfig = get_env_kubectl(env_name)

    if not kubeconfig:
        print("[red]Env kubeconfig not found")
        return

    env_kubeconfig_path = get_config_path() / f"kubeconfig-{env_name}"
    with open(env_kubeconfig_path, "w") as kubeconfig_file:
        kubeconfig_file.write(kubeconfig)

    print(f"export KUBECONFIG={env_kubeconfig_path}")
    return


@app.command(name="monitor")
async def monitor_info_cmd(
    *, env: Annotated[str, Parameter(name=["-e", "--env"], required=True)]
):
    env_name = resolve_env_alias(env)
    monitor_info = get_env_monitor_info(env_name)
    print(monitor_info.url)


# TODO: Add validation command to check repo, azure resources?, cluster status?


@app.command(name="version")
async def env_version_cmd(env_name: str):
    try:
        version = await load_env_version(env_name)
        print(version)
    except Exception as e:
        print(e)


@app.command(name="status")
async def env_status_cmd(env_name: str):
    try:
        env_desc = await env_desc_loader.load_env_desc(env_name)

        if env_desc is None:
            print("Env not found")
            return
    except Exception as e:
        print(e)
        return

    env_client = EnvClient(env_desc)
    resource_clients = env_client.get_resource_clients()

    with Progress(
        SpinnerColumn(),
        TextColumn("{task.description}"),
        TextColumn("{task.fields[result]}"),
    ) as table:
        loop = asyncio.get_event_loop()
        tasks = [
            (
                resource_client,
                loop.create_task(_get_formatted_status(resource_client.client)),
                table.add_task(f"[blue]{resource_client.name}[/blue]", result=""),
            )
            for resource_client in resource_clients
        ]

        while not table.finished:
            for resource_client, task, table_task in tasks:
                if task.done():
                    table.update(
                        table_task, total=100, completed=100, result=task.result()
                    )

            await asyncio.sleep(0.1)

    [await resource_client.client.close() for resource_client in resource_clients]


@app.command(name="upgrade")
async def upgrade_env_cmd(env_name: str, to_version: str):
    await migrate_env(env_name, to_version)


async def _get_formatted_status(resource_client):
    status = await resource_client.status()
    return print_status(status)



from rich import print

from btpy.core.envs.env_desc_loader import load_env_desc
from btpy.core.migrations.migration_loader import (
    find_migrations,
    to_pretty_migration_name,
)
from btpy.core.migrations.migration_runner import run_migration


async def migrate_env(env_name: str, to_version: str):
    try:
        env_desc = await load_env_desc(env_name)

        if env_desc.version == to_version:
            print(f"[yellow]Env {env_name} already at version {to_version}")
            return

        if env_desc.version > to_version:
            print(
                f"[red]New version must be higher than current version ({env_desc.version})"
            )
            return

        required_migrations = await find_migrations(env_desc.version, to_version)

        if len(required_migrations) == 0:
            print(
                f"[cyan]No migrations required for upgrade from {env_desc.version} to {to_version}"
            )
        else:
            print(
                f"[cyan]Found {len(required_migrations)} migrations required for upgrade from {env_desc.version} to {to_version}"
            )

        for migration in required_migrations:
            print(f"[cyan]Running migration: {to_pretty_migration_name(migration)}")
            migration_result = await run_migration(env_desc, migration)

            if not migration_result:
                print("[red]Migration failed")
                return
            else:
                print("[green]Migration done")
    except Exception as e:
        print(f"[red]ERROR: {e}")



import json
from dataclasses import asdict, dataclass, field

from btpy.core import rancher

from btpy.configuration.config import KeyVaultInfo, get_config_value
from btpy.core.azure_utility import get_keyvault_secret
from btpy.core.misc_utility import get_or_create_azure_tenant_sp


@dataclass
class RancherManagementConfig:
    url: str


@dataclass
class RancherCredential:
    client_id: str
    client_secret: str
    subscription_id: str
    tenant_id: str
    name: str


@dataclass
class RancherKeys:
    access_key: str
    secret_key: str


@dataclass
class AzureCredentialConfig:
    clientId: str
    clientSecret: str
    subscriptionId: str
    tenantId: str


@dataclass
class RancherCloudCredentialRequest:
    annotations: dict[str, str]
    name: str
    azurecredentialConfig: AzureCredentialConfig


def get_or_create_rancher_azure_credential(
    tenant_id, subscription_id
) -> RancherCredential:
    client = _get_rancher_client()
    cloud_credentials = client.list_cloud_credential()["data"]
    rancher_cred = None

    for cred in cloud_credentials:
        if cred["annotations"]["provisioning.cattle.io/driver"] != "azure":
            continue
        if cred["azurecredentialConfig"]["tenantId"] != tenant_id:
            continue
        if cred["azurecredentialConfig"]["subscriptionId"] != subscription_id:
            continue

        az_sp = get_or_create_azure_tenant_sp(
            cred["azurecredentialConfig"]["tenantId"],
            cred["azurecredentialConfig"]["subscriptionId"],
        )
        rancher_cred = RancherCredential(
            client_id=cred["azurecredentialConfig"]["clientId"],
            client_secret=az_sp.client_secret,
            subscription_id=cred["azurecredentialConfig"]["subscriptionId"],
            tenant_id=cred["azurecredentialConfig"]["tenantId"],
            name=cred["name"],
        )
        break
    if not rancher_cred:
        az_sp = get_or_create_azure_tenant_sp(tenant_id, subscription_id)
        cred_config = RancherCloudCredentialRequest(
            annotations={"provisioning.cattle.io/driver": "azure"},
            name=f"az-tenant-{tenant_id}-{subscription_id}",
            azurecredentialConfig=AzureCredentialConfig(
                clientId=az_sp.client_id,
                clientSecret=az_sp.client_secret,
                subscriptionId=subscription_id,
                tenantId=tenant_id,
            ),
        )
        created_cred = client.create_cloud_credential(asdict(cred_config))
        rancher_cred = RancherCredential(
            client_id=created_cred["azurecredentialConfig"]["clientId"],
            client_secret=az_sp.client_secret,
            subscription_id=created_cred["azurecredentialConfig"]["subscriptionId"],
            tenant_id=created_cred["azurecredentialConfig"]["tenantId"],
            name=created_cred["name"],
        )

    return rancher_cred


def _get_rancher_client():
    mgmt_info = get_rancher_mgmt_info()
    return rancher.Client(
        url=mgmt_info.url,
        access_key=mgmt_info.access_key,
        secret_key=mgmt_info.secret_key,
    )


@dataclass
class RancherMgmtInfo:
    url: str
    access_key: str
    secret_key: str
    token: str = field(init=False)

    def __post_init__(self):
        self.token = f"{self.access_key}:{self.secret_key}"


def get_rancher_mgmt_info() -> RancherMgmtInfo:
    rancher_mgmt_config = RancherManagementConfig(**get_config_value("rancher_mgmt"))
    if not rancher_mgmt_config:
        raise Exception("[red]Rancher management config not found in btpy config")

    kv_info = KeyVaultInfo(**get_config_value("global_keyvault"))
    rancher_keys = RancherKeys(
        **json.loads(get_keyvault_secret(kv_info.name, "rancher-mgmt-keys"))
    )
    if not rancher_keys:
        raise Exception("[red]Rancher management keys not found in global key vault")

    return RancherMgmtInfo(
        url=rancher_mgmt_config.url,
        access_key=rancher_keys.access_key,
        secret_key=rancher_keys.secret_key,
    )



from pathlib import Path

import yaml

from .config import get_config_path


def get_aliases_file_path(alias_type: str = "env") -> Path:
    """Get the path to the aliases file.

    Args:
        alias_type: Type of aliases ('env' or 'tenant')
    """
    config_dir = get_config_path()
    if alias_type == "tenant":
        return config_dir / "tenant_aliases.yaml"
    return config_dir / "env_aliases.yaml"


def load_aliases(alias_type: str = "env") -> dict[str, str]:
    """Load aliases from the aliases file.

    Args:
        alias_type: Type of aliases ('env' or 'tenant')

    Returns:
        Dictionary mapping alias -> name
    """
    aliases_path = get_aliases_file_path(alias_type)

    if not aliases_path.exists():
        return {}

    with open(aliases_path, "r") as f:
        aliases = yaml.safe_load(f)
        return aliases if aliases else {}


def save_aliases(aliases: dict[str, str], alias_type: str = "env") -> None:
    """Save aliases to the aliases file.

    Args:
        aliases: Dictionary mapping alias -> name
        alias_type: Type of aliases ('env' or 'tenant')
    """
    aliases_path = get_aliases_file_path(alias_type)

    # Ensure config directory exists
    aliases_path.parent.mkdir(parents=True, exist_ok=True)

    with open(aliases_path, "w") as f:
        yaml.dump(aliases, f, default_flow_style=False, sort_keys=True)


def resolve_env_alias(name_or_alias: str) -> str:
    """Resolve an alias to its environment name, or return the input if not an alias.

    Args:
        name_or_alias: Either an alias or an actual environment name

    Returns:
        The resolved environment name
    """
    aliases = load_aliases("env")
    return aliases.get(name_or_alias, name_or_alias)


def resolve_tenant_alias(name_or_alias: str) -> str:
    """Resolve an alias to its tenant name, or return the input if not an alias.

    Args:
        name_or_alias: Either an alias or an actual tenant name

    Returns:
        The resolved tenant name
    """
    aliases = load_aliases("tenant")
    return aliases.get(name_or_alias, name_or_alias)


def set_alias(alias: str, name: str, alias_type: str = "env") -> None:
    """Set an alias for a name.

    Args:
        alias: The alias to set
        name: The actual name
        alias_type: Type of aliases ('env' or 'tenant')
    """
    aliases = load_aliases(alias_type)
    aliases[alias] = name
    save_aliases(aliases, alias_type)


def remove_alias(alias: str, alias_type: str = "env") -> bool:
    """Remove an alias.

    Args:
        alias: The alias to remove
        alias_type: Type of aliases ('env' or 'tenant')

    Returns:
        True if alias was removed, False if it didn't exist
    """
    aliases = load_aliases(alias_type)
    if alias in aliases:
        del aliases[alias]
        save_aliases(aliases, alias_type)
        return True
    return False



import shlex
import time
from dataclasses import dataclass
from typing import Optional

from azure.identity import AzureCliCredential, DefaultAzureCredential
from azure.mgmt.resource import ResourceManagementClient
from azure.storage.blob import BlobServiceClient, ContainerClient
from msgraph import GraphServiceClient
from msgraph.generated.models.application import Application
from msgraph.generated.models.password_credential import PasswordCredential
from msgraph.generated.models.web_application import WebApplication
from rich import print

from btpy.az import az
from btpy.configuration.config import get_env_data_storage_name


@dataclass
class AzureADInfo:
    """Azure AD B2C tenant information"""

    tenant_id: str
    domain_name: str
    resource_id: str
    location: str
    billing_type: Optional[str]
    sku: Optional[str]
    properties: dict


@dataclass
class AzureADAppInfo:
    """Azure AD app registration information"""

    id: str  # Object ID
    app_id: str  # Application (client) ID
    display_name: str
    sign_in_audience: Optional[str]
    redirect_uris: list[str]


@dataclass
class AzureADAppSecretInfo:
    """Azure AD app registration secret information"""

    key_id: str
    display_name: str
    start_date_time: Optional[str]
    end_date_time: Optional[str]
    hint: Optional[str]  # Last few characters of the secret
    secret_text: Optional[str] = None  # Only available at creation time


_blob_service_client = None


def get_azure_credential():
    return AzureCliCredential()
    # return DefaultAzureCredential()


def to_azure_appsetting_format(setting_value: str) -> str:
    return setting_value.replace(":", "__")


def from_azure_appsetting_format(setting_value: str) -> str:
    return setting_value.replace("__", ":")


def get_blob_service_client() -> BlobServiceClient:
    global _blob_service_client

    if _blob_service_client is None:
        credential = get_azure_credential()
        blob_account_url = (
            f"https://{get_env_data_storage_name()}.blob.core.windows.net"
        )
        _blob_service_client = BlobServiceClient(
            blob_account_url, credential=credential
        )

    return _blob_service_client


def get_blob_container_client(container_name: str) -> ContainerClient:
    blob_service_client = get_blob_service_client()
    return blob_service_client.get_container_client(container_name)


def get_blob_client(container_name: str, blob_name: str):
    container_client = get_blob_container_client(container_name)
    return container_client.get_blob_client(blob_name)


def to_blob_index_format(tag: str) -> str:
    return tag.replace("-", "_")


def storage_account_exists(storage_account_name):
    exit_code, result, err = az(f"storage account check-name -n {storage_account_name}")

    if exit_code != 0:
        print(
            f"[red]Failed to check if Azure storage account exists ({exit_code} - {err})"
        )
        raise Exception("Failed to check if Azure storage account exists")

    return not result["nameAvailable"]


def resource_group_exists(resource_group_name):
    exit_code, result, err = az(f"group exists -n {resource_group_name}")

    if exit_code != 0:
        print(
            f"[red]Failed to check if Azure resource group exists ({exit_code} - {err})"
        )
        raise Exception("Failed to check if Azure resource group exists")

    return result


def create_storage_account(resource_group_name, storage_account_name, location):
    if storage_account_exists(storage_account_name):
        print(f"[blue]Storage account {storage_account_name} already exists")
        return

    print(f"[blue]Creating Azure storage account {storage_account_name}...", end=" ")
    exit_code, result, err = az(
        f"storage account create -g {resource_group_name} -n {storage_account_name} --sku Standard_LRS -l {location}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception("Failed to create Azure storage account")

    print("[green]Complete")


def list_storage_table_entries(storage_account_name, table_name, subscription_id=None):
    exit_code, result, err = az(
        f"storage entity query --account-name {storage_account_name} --table-name {table_name}",
        subscription_id,
    )

    if exit_code != 0:
        raise Exception(
            f"Failed to list Azure table storage table entries ({exit_code} - {err})"
        )

    return result


def get_storage_table_entry(
    storage_account_name, table_name, subscription_id, partition_key, row_key
):
    exit_code, result, err = az(
        f"storage entity show --account-name {storage_account_name} --table-name {table_name} --partition-key {partition_key} --row-key {row_key}",
        subscription_id,
    )

    if exit_code != 0:
        if "The specified resource does not exist." in err:
            return None
        raise Exception(
            f"Failed to get Azure table storage table entry ({exit_code} - {err})"
        )

    return result


def upsert_table_storage_entry(
    storage_account_name,
    table_name,
    partition_key,
    row_key,
    data,
    subscription_id=None,
    log=True,
):
    if log:
        print(
            f"[blue]Upserting Azure table storage entity {partition_key}-{row_key}...",
            end=" ",
        )
    exit_code, result, err = az(
        f"storage entity insert --table-name {table_name} --account-name {storage_account_name} --entity PartitionKey='{partition_key}' RowKey='{row_key}' Data='{data}' --if-exists replace",
        subscription_id,
    )

    if exit_code != 0:
        if log:
            print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to upsert Azure table storage table entry ({exit_code} - {err})"
        )

    if log:
        print("[green]Complete")


def create_resource_group(resource_group_name, location):
    if resource_group_exists(resource_group_name):
        print(f"[blue]Resource group {resource_group_name} already exists")
        return

    print(f"[blue]Creating Azure resource group {resource_group_name}...", end=" ")
    exit_code, result, err = az(f"group create -n {resource_group_name} -l {location}")

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception("Failed to create Azure resource group")

    print("[green]Complete")


def container_exists(storage_connection_string, container):
    exit_code, result, err = az(
        f'storage container exists --connection-string "{storage_connection_string}" --name {container} --query exists'
    )

    if exit_code != 0:
        print(
            f"[red]Failed to check if blob container {container} exists ({exit_code} - {err})"
        )
        raise Exception("Failed to check if Azure blob container exists")

    return result


def create_container(storage_connection_string, container):
    if container_exists(storage_connection_string, container):
        print(f"[blue]Blob container {container} already exists")
        return None

    print(f"[blue]Creating Azure blob container {container}...", end=" ")
    exit_code, result, err = az(
        f"storage container create --connection-string {storage_connection_string} --name {container}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception("Failed to create Azure blob container")

    print("[green]Complete")

    return result


def get_storage_connection_string(storage_name):
    exit_code, result, err = az(
        f"storage account show-connection-string --name {storage_name} --query connectionString"
    )

    if exit_code != 0:
        print(
            f"[red]Failed to get Azure storage account connection string ({exit_code} - {err})"
        )
        raise Exception("Failed to get Azure storage account connection string")

    return result


def blob_exists(storage_connection_string, container, blob_name):
    exit_code, result, err = az(
        f'storage blob exists --connection-string "{storage_connection_string}" --container-name {container} --name {blob_name} --query exists'
    )

    if exit_code != 0:
        print(f"[red]Failed to check if Azure blob exists ({exit_code} - {err})")
        raise Exception("Failed to check if Azure blob exists")

    return result


def get_current_azure_user_id():
    exit_code, result, err = az("ad signed-in-user show --query id -o tsv")

    if exit_code != 0:
        raise Exception(f"Failed to get current user ID ({exit_code} - {err})")

    return result


def assign_keyvault_reader_role(kv_id, user_id, subscription_id):
    name = kv_id.split("/")[-1]
    print(
        f"[blue]Assigning secret user role to {user_id} for Azure key vault {name}...",
        end=" ",
    )
    exit_code, result, err = az(
        f'role assignment create --role "Key Vault Secrets User" --assignee {user_id} --scope {kv_id}',
        subscription_id,
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to assign key vault secret user role to {user_id} for Azure key vault {name}"
        )

    print("[green]Complete")
    return result


def set_keyvault_admin_rbac(kv_id, user_id):
    name = kv_id.split("/")[-1]
    print(
        f"[blue]Assigning admin role to {user_id} for Azure key vault {name}...",
        end=" ",
    )
    exit_code, result, err = az(
        f'role assignment create --role "Key Vault Administrator" --assignee {user_id} --scope {kv_id}'
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to assign admin role to {user_id} for Azure key vault {name}"
        )

    print("[green]Complete")
    return result


def assign_subscription_role(subscription_id, assignee_id, role):
    print(
        f"[blue]Assigning {role} role to {assignee_id} for subscription {subscription_id}...",
        end=" ",
    )
    scope = f"/subscriptions/{subscription_id}"
    exit_code, result, err = az(
        f'role assignment create --role "{role}" --assignee {assignee_id} --scope {scope}'
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to assign {role} role to {assignee_id} for subscription {subscription_id} ({exit_code} - {err})"
        )

    print("[green]Complete")
    return result


def is_keyvault_admin(kv_id, user_id):
    exit_code, result, err = az(f'role assignment list --scope "{kv_id}"')

    if exit_code != 0:
        raise Exception(
            f"Failed to get Azure key vault admin list ({exit_code} - {err})"
        )

    for admin in result:
        if (
            admin["principalId"] == user_id
            and admin["roleDefinitionName"] == "Key Vault Administrator"
        ):
            return True

    return False


def await_keyvault_admin_assignment(kv_id, user_id, timeout_s=60):
    poll_timer_s = 2
    start_time = time.time()

    while time.time() - start_time < timeout_s:
        time.sleep(poll_timer_s)
        if is_keyvault_admin(kv_id, user_id):
            return True

    return False


def keyvault_id(resource_group, name):
    exit_code, result, err = az(
        f"keyvault show --name {name} --resource-group {resource_group} --query id -o tsv"
    )
    return result


def keyvault_exists(resource_group, name):
    exit_code, result, err = az(
        f"keyvault show --name {name} --resource-group {resource_group} --query id -o tsv"
    )
    return exit_code == 0


def get_keyvault(resource_group, name):
    exit_code, result, err = az(
        f"keyvault show --name {name} --resource-group {resource_group}"
    )

    if exit_code != 0:
        raise Exception("Failed to get Azure key vault")

    return result


def create_keyvault(resource_group, name, location):
    if keyvault_exists(resource_group, name):
        print(f"[blue]Azure key vault {name} already exists")
        return get_keyvault(resource_group, name)

    print(f"[blue]Creating Azure key vault {name}...", end=" ")
    exit_code, result, err = az(
        f'keyvault create --name "{name}" --resource-group "{resource_group}" --location "{location}" --sku standard'
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception("Failed to create Azure key vault")

    print("[green]Complete")

    id = result["id"]
    user_id = get_current_azure_user_id()
    set_keyvault_admin_rbac(id, user_id)

    return result


def get_keyvault_id(kv_name):
    exit_code, result, err = az(f"keyvault show --name {kv_name} --query id -o tsv")

    if exit_code != 0:
        print(f"[red]Failed to get Azure key vault id ({exit_code} - {err})")
        raise Exception("Failed to get Azure key vault id")

    return result


def get_keyvault_secret(kv_name, secret_name):
    exit_code, result, err = az(
        f"keyvault secret show --vault-name {kv_name} --name {secret_name} --query value -o tsv"
    )

    if exit_code != 0:
        if "was not found in this key vault" in err:
            return None
        print(f"[red]Failed to get Azure key vault secret value ({exit_code} - {err})")
        raise Exception(
            f"Failed to get Azure key vault secret value ({exit_code} - {err})"
        )

    return result


def get_keyvault_secret_to_file(kv_name, secret_name, path):
    secret_value = get_keyvault_secret(kv_name, secret_name)

    with open(path, "w") as f:
        f.write(secret_value)


def query_keyvault_secrets(kv_name, jmespath_query, subscription_id=None):
    exit_code, result, err = az(
        f'keyvault secret list --vault-name {kv_name} --query "{jmespath_query}"',
        subscription_id,
    )

    if exit_code != 0:
        raise Exception(
            f"Failed to query Azure key vault secrets ({exit_code} - {err})"
        )

    return result


def upload_keyvault_secret_file(kv_name, secret_name, file_path):
    print(f"[blue]Uploading Azure key vault secret {secret_name} from file...", end=" ")
    exit_code, result, err = az(
        f"keyvault secret set --vault-name {kv_name} --name {secret_name} --file {file_path}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to upload Azure key vault secret from file ({exit_code} - {err})"
        )

    print("[green]Complete")
    return result


def upload_keyvault_secret(kv_name, secret_name, secret_value):
    print(f"[blue]Uploading Azure key vault secret {secret_name}...", end=" ")
    exit_code, result, err = az(
        f"keyvault secret set --vault-name {kv_name} --name {secret_name} --value {shlex.quote(secret_value)}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to upload Azure key vault secret ({exit_code} - {err})"
        )

    print("[green]Complete")
    return result


def delete_keyvault_secret(kv_name, secret_name):
    print(f"[blue]Deleting Azure key vault secret {secret_name}...", end=" ")
    exit_code, result, err = az(
        f"keyvault secret delete --vault-name {kv_name} --name {secret_name}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to delete Azure key vault secret ({exit_code} - {err})"
        )

    exit_code, result, err = az(
        f"keyvault secret purge --vault-name {kv_name} --name {secret_name}"
    )

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(f"Failed to purge Azure key vault secret ({exit_code} - {err})")

    print("[green]Complete")


def get_azure_service_principal(env_name):
    sp_name = f"sp-bt-env-{env_name}"
    exit_code, result, err = az(f'ad sp list --display-name "{sp_name}"')

    if exit_code != 0:
        print(
            f"[red]Failed to get Azure service principal exists ({exit_code} - {err})"
        )
        raise Exception(
            f"Failed to get Azure service principal exists ({exit_code} - {err})"
        )

    return result[0] if len(result) > 0 else None


def azure_service_principal_exists(env_name):
    sp_name = f"sp-bt-env-{env_name}"
    exit_code, result, err = az(f'ad sp list --display-name "{sp_name}"')

    if exit_code != 0:
        print(
            f"[red]Failed to check if Azure service principal exists ({exit_code} - {err})"
        )
        raise Exception(
            f"Failed to check if Azure service principal exists ({exit_code} - {err})"
        )

    return len(result) > 0


def validate_current_tenant(expected_tenant_id):
    """Validate that the current Azure CLI context matches the expected tenant.

    Raises an exception if there's a mismatch.
    """
    exit_code, result, err = az("account show")
    if exit_code != 0:
        raise Exception(f"Failed to get current Azure account ({exit_code} - {err})")

    current_tenant = result.get("tenantId")
    if current_tenant != expected_tenant_id:
        raise Exception(
            f"Azure CLI tenant mismatch!\n"
            f"  Expected tenant: {expected_tenant_id}\n"
            f"  Current tenant:  {current_tenant}\n"
            f"  Please switch tenant with: az login --tenant {expected_tenant_id}"
        )


def validate_current_subscription(expected_subscription_id):
    """Validate that the current Azure CLI context matches the expected subscription.

    Raises an exception if there's a mismatch.
    """
    exit_code, result, err = az("account show")
    if exit_code != 0:
        raise Exception(f"Failed to get current Azure account ({exit_code} - {err})")

    current_subscription = result.get("id")
    if current_subscription != expected_subscription_id:
        raise Exception(
            f"Azure CLI subscription mismatch!\n"
            f"  Expected subscription: {expected_subscription_id}\n"
            f"  Current subscription:  {current_subscription}\n"
            f"  Please switch subscription with: az account set --subscription {expected_subscription_id}"
        )


def create_service_principal(name, tenant_id=None):
    print("[blue]Creating Azure service principal...", end=" ")

    if tenant_id:
        validate_current_tenant(tenant_id)

    exit_code, result, err = az(f'ad sp create-for-rbac --name "{name}"')

    if exit_code != 0:
        print(f"[red]Failed ({exit_code} - {err})")
        raise Exception(
            f"Failed to create Azure service principal ({exit_code} - {err})"
        )

    print("[green]Complete")
    return result


def get_service_principals_by_name(name):
    exit_code, result, err = az(f"""ad sp list --filter "displayname eq '{name}'" """)

    if exit_code != 0:
        raise Exception(
            f"Failed to list Azure service principals ({exit_code} - {err})"
        )

    return result


def delete_service_principal(id):
    exit_code, result, err = az(f"ad sp delete --id {id}")

    if exit_code != 0:
        raise Exception(
            f"Failed to delete Azure service principal ({exit_code} - {err})"
        )


def get_current_subscription():
    exit_code, result, err = az("account show")

    if exit_code != 0:
        raise Exception(f"Failed to get Azure tenant ID ({exit_code} - {err})")

    return result


def get_azure_tenant_id():
    exit_code, result, err = az("account show --query tenantId -o tsv")

    if exit_code != 0:
        raise Exception(f"Failed to get Azure tenant ID ({exit_code} - {err})")

    return result


def create_azure_ad(
    id, name, resource_group, location, country_code, subscription_id
) -> AzureADInfo:
    """
    Create an Azure Active Directory B2C tenant within the specified resource group in the provided
    Azure subscription. This function initializes a `ResourceManagementClient` using Azure's default
    credentials to manage the Azure resources.

    Parameters
    ----------
    id : str
        The unique identifier for the Azure Active Directory B2C tenant. E.g. btglobalneu1234.
    name : str
        The display name for the Azure Active Directory B2C tenant. E.g. Glaze.
    resource_group : str
        The name of the Azure resource group within which the directory will be created.
    location : str
        The geographic location for the Azure Active Directory B2C tenant. NOTE: These are not the standard Azure location names! Valid values are: "United States", "Europe", "Asia Pacific" and "Australia".
    country_code: str
        The country code within the location specified. Regular country codes. Used for billing etc. E.g. "SE".
    subscription_id : str
        The Azure subscription ID under which the resource group and the tenant reside.

    Returns
    -------
    AzureADInfo
        Information about the created Azure AD B2C tenant.
    """
    credential = DefaultAzureCredential()
    client = ResourceManagementClient(credential, subscription_id)
    tenant_domain = f"{id}.onmicrosoft.com"

    result = client.resources.begin_create_or_update(
        resource_group_name=resource_group,
        resource_provider_namespace="Microsoft.AzureActiveDirectory",
        parent_resource_path="",
        resource_type="b2cDirectories",
        resource_name=tenant_domain,
        api_version="2021-04-01",
        parameters={
            "location": location,
            "sku": {"name": "Standard", "tier": "A0"},
            "properties": {
                "createTenantProperties": {
                    "displayName": name,
                    "countryCode": country_code,
                }
            },
        },
    ).result()

    # Construct AzureADInfo from creation result
    return AzureADInfo(
        tenant_id=result.properties.get("tenantId"),
        domain_name=result.name,
        resource_id=result.id,
        location=result.location,
        billing_type=result.properties.get("billingConfig", {}).get("billingType"),
        sku=result.sku.name if result.sku else None,
        properties=result.properties,
    )


def get_azure_ad(
    tenant_domain, resource_group, subscription_id
) -> Optional[AzureADInfo]:
    credential = DefaultAzureCredential()
    client = ResourceManagementClient(credential, subscription_id)

    try:
        resource = client.resources.get(
            resource_group_name=resource_group,
            resource_provider_namespace="Microsoft.AzureActiveDirectory",
            parent_resource_path="",
            resource_type="b2cDirectories",
            resource_name=tenant_domain,
            api_version="2021-04-01",
        )

        # Extract useful info
        return AzureADInfo(
            tenant_id=resource.properties.get("tenantId"),
            domain_name=resource.name,
            resource_id=resource.id,
            location=resource.location,
            billing_type=resource.properties.get("billingConfig", {}).get(
                "billingType"
            ),
            sku=resource.sku.name if resource.sku else None,
            properties=resource.properties,
        )
    except Exception as e:
        print(f"WTF BRO ({e})")
        return None


def _app_to_info(app) -> AzureADAppInfo:
    """Convert Graph SDK Application object to AzureADAppInfo."""
    redirect_uris = []
    if app.web and app.web.redirect_uris:
        redirect_uris = app.web.redirect_uris
    return AzureADAppInfo(
        id=app.id,
        app_id=app.app_id,
        display_name=app.display_name,
        sign_in_audience=app.sign_in_audience,
        redirect_uris=redirect_uris,
    )


def _secret_to_info(cred, secret_text=None) -> AzureADAppSecretInfo:
    """Convert Graph SDK PasswordCredential to AzureADAppSecretInfo."""
    return AzureADAppSecretInfo(
        key_id=str(cred.key_id) if cred.key_id else "",
        display_name=cred.display_name or "",
        start_date_time=cred.start_date_time.isoformat()
        if cred.start_date_time
        else None,
        end_date_time=cred.end_date_time.isoformat() if cred.end_date_time else None,
        hint=cred.hint,
        secret_text=secret_text or cred.secret_text,
    )


async def create_azure_ad_app(name, redirect_uris, tenant_id) -> AzureADAppInfo:
    from msgraph.generated.models.implicit_grant_settings import ImplicitGrantSettings

    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )
    app_desc = Application(
        display_name=name,
        sign_in_audience="AzureADandPersonalMicrosoftAccount",
        web=WebApplication(
            redirect_uris=redirect_uris,
            implicit_grant_settings=ImplicitGrantSettings(
                enable_id_token_issuance=True,
                enable_access_token_issuance=True,
            ),
        ),
    )
    app = await client.applications.post(app_desc)
    return _app_to_info(app)


async def azure_ad_app_exists(name, tenant_id) -> bool:
    return await get_azure_ad_app_by_name(name, tenant_id) is not None


async def get_azure_ad_app_by_name(name, tenant_id) -> Optional[AzureADAppInfo]:
    from msgraph.generated.applications.applications_request_builder import (
        ApplicationsRequestBuilder,
    )

    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )

    query_params = (
        ApplicationsRequestBuilder.ApplicationsRequestBuilderGetQueryParameters(
            filter=f"displayName eq '{name}'"
        )
    )
    config = (
        ApplicationsRequestBuilder.ApplicationsRequestBuilderGetRequestConfiguration(
            query_parameters=query_params
        )
    )
    apps = await client.applications.get(request_configuration=config)

    if apps and apps.value and len(apps.value) > 0:
        return _app_to_info(apps.value[0])
    return None


async def get_azure_ad_apps(tenant_id) -> list[AzureADAppInfo]:
    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )
    result = await client.applications.get()
    if result and result.value:
        return [_app_to_info(app) for app in result.value]
    return []


async def get_azure_ad_app(app_id, tenant_id) -> Optional[AzureADAppInfo]:
    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )
    app = await client.applications.by_application_id(app_id).get()
    if app:
        return _app_to_info(app)
    return None


async def create_azure_ad_app_secret(
    app_id, secret_name, tenant_id
) -> AzureADAppSecretInfo:
    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )
    secret_desc = PasswordCredential(display_name=secret_name)
    cred = await client.applications.by_application_id(app_id).add_password.post(
        secret_desc
    )
    return _secret_to_info(cred)


async def get_azure_ad_app_secret_by_name(
    app_id, secret_name, tenant_id
) -> Optional[AzureADAppSecretInfo]:
    """Get app secret metadata by display name.

    Note: The actual secret value is only available at creation time.
    This returns metadata (key_id, display_name, end_date_time, etc.)
    """
    credential = DefaultAzureCredential(additionally_allowed_tenants=[tenant_id])
    client = GraphServiceClient(
        credentials=credential, scopes=["https://graph.microsoft.com/.default"]
    )
    app = await client.applications.by_application_id(app_id).get()
    if app and app.password_credentials:
        for cred in app.password_credentials:
            if cred.display_name == secret_name:
                return _secret_to_info(cred)
    return None


def listen_to_eventhub(connection_string):
    return


def location_tag(location):
    match location.lower():
        case "australiacentral":
            return "auc"
        case "australiacentral2":
            return "auc2"
        case "australiaeast":
            return "aue"
        case "australiasoutheast":
            return "ause"
        case "austriaeast":
            return "ate"
        case "belgiumcentral":
            return "bec"
        case "brazilsouth":
            return "brs"
        case "brazilsoutheast":
            return "brse"
        case "canadacentral":
            return "cac"
        case "canadaeast":
            return "cae"
        case "centralindia":
            return "cin"
        case "centralus":
            return "cus"
        case "chilecentral":
            return "clc"
        case "eastasia":
            return "eas"
        case "eastus":
            return "eus"
        case "eastus2":
            return "eus2"
        case "francecentral":
            return "frc"
        case "francesouth":
            return "frs"
        case "germanynorth":
            return "gen"
        case "germanywestcentral":
            return "gewc"
        case "indonesiacentral":
            return "idc"
        case "israelcentral":
            return "ilc"
        case "italynorth":
            return "itn"
        case "japaneast":
            return "jpe"
        case "japanwest":
            return "jpw"
        case "koreacentral":
            return "krc"
        case "koreasouth":
            return "krs"
        case "malaysiawest":
            return "myw"
        case "mexicocentral":
            return "mxc"
        case "newzealandnorth":
            return "nzn"
        case "northcentralus":
            return "ncus"
        case "northeurope":
            return "neu"
        case "norwayeast":
            return "noe"
        case "norwaywest":
            return "now"
        case "polandcentral":
            return "plc"
        case "qatarcentral":
            return "qac"
        case "southafricanorth":
            return "zan"
        case "southafricawest":
            return "zaw"
        case "southcentralus":
            return "scus"
        case "southindia":
            return "sin"
        case "southeastasia":
            return "seas"
        case "spaincentral":
            return "esc"
        case "swedencentral":
            return "sec"
        case "switzerlandnorth":
            return "chn"
        case "switzerlandwest":
            return "chw"
        case "uaecentral":
            return "aec"
        case "uaenorth":
            return "aen"
        case "uksouth":
            return "uks"
        case "ukwest":
            return "ukw"
        case "westcentralus":
            return "wcus"
        case "westeurope":
            return "weu"
        case "westindia":
            return "win"
        case "westus":
            return "wus"
        case "westus2":
            return "wus2"
        case "westus3":
            return "wus3"
        case _:
            raise ValueError(f"Unexpected Azure location: {location}")


if __name__ == "__main__":
    keyvault_exists("rg-bt-env-glaze-env3-neu6vyx", "kvbtglazeenv3neu6vyx")
